# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node visitor as defined in https://docs.python.org/3/library/ast.html#ast.NodeVisitor
This module is used internally by ``qiskit.synthesis.boolean.BooleanExpression``.
"""

import ast
import _ast


class BooleanExpressionEvalVisitor(ast.NodeVisitor):
    """Node visitor to compute the value of the expression, given the boolean values of the args
    as defined in https://docs.python.org/3/library/ast.html#ast.NodeVisitor"""

    # pylint: disable=invalid-name
    bitops = {
        _ast.BitAnd: lambda values: values[0] and values[1],
        _ast.And: lambda values: values[0] and values[1],
        _ast.BitOr: lambda values: values[0] or values[1],
        _ast.Or: lambda values: values[0] or values[1],
        _ast.BitXor: lambda values: values[0] ^ values[1],
        _ast.Not: lambda values: not values[0],
        _ast.Invert: lambda values: not values[0],
    }

    def __init__(self):
        self.arg_values = {}
        super().__init__()

    def bit_binop(self, op, values):
        """Performs the operation, if it is recognized"""
        op_type = type(op)
        if op_type not in self.bitops:
            raise ValueError(f"Unknown op: {op_type}")
        return self.bitops[op_type](values)

    def visit_BinOp(self, node):
        """Handles ``&``, ``^``, and ``|``."""
        return self.bit_binop(node.op, [self.visit(node.left), self.visit(node.right)])

    def visit_UnaryOp(self, node):
        """Handles ``~``."""
        return self.bit_binop(node.op, [self.visit(node.operand)])

    def visit_Name(self, node):
        """Reduce variable names."""
        if node.id not in self.arg_values:
            raise ValueError(f"Undefined value for {node.id}")
        return self.arg_values[node.id]

    def visit_Module(self, node):
        """Returns the value of the single expression comprising the boolean expression"""
        if len(node.body) != 1 or not isinstance(node.body[0], ast.Expr):
            raise ValueError("Incorrectly formatted boolean expression")
        return self.visit(node.body[0])

    def visit_Expr(self, node):
        """Returns the value of the expression"""
        return self.visit(node.value)

    def generic_visit(self, node):
        """Catch all for the unhandled nodes."""
        raise ValueError(f"Unknown node: {type(node)}")


class BooleanExpressionArgsCollectorVisitor(ast.NodeVisitor):
    """Node visitor to collect the name of the args of the expression
    as defined in https://docs.python.org/3/library/ast.html#ast.NodeVisitor"""

    # pylint: disable=invalid-name
    def __init__(self):
        self.args = set()
        self.args_pos = {}
        super().__init__()

    def visit_Name(self, node):
        """Collect arg name."""
        self.args.add(node.id)
        if node.id not in self.args_pos or (
            self.args_pos[node.id] > (node.lineno, node.col_offset)
        ):
            self.args_pos[node.id] = (node.lineno, node.col_offset)

    def get_sorted_args(self):
        """Returns a list of the args, sorted by their appearance locations"""
        return sorted(self.args, key=lambda arg: self.args_pos[arg])



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022, 2023, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Base Estimator V1 and V2 classes"""

from __future__ import annotations

from abc import abstractmethod, ABC
from collections.abc import Iterable, Sequence
from copy import copy
from typing import Generic, TypeVar

from qiskit.circuit import QuantumCircuit
from qiskit.providers import JobV1 as Job
from qiskit.quantum_info.operators import SparsePauliOp
from qiskit.quantum_info.operators.base_operator import BaseOperator

from ..containers import (
    DataBin,
    EstimatorPubLike,
    PrimitiveResult,
    PubResult,
)
from ..containers.estimator_pub import EstimatorPub
from .validation_v1 import _validate_estimator_args
from .base_primitive_v1 import BasePrimitiveV1
from .base_primitive_job import BasePrimitiveJob

T = TypeVar("T", bound=Job)


class BaseEstimatorV2(ABC):
    r"""Base class for ``EstimatorV2`` implementations.

    An estimator calculates expectation values for provided quantum circuit and
    observable combinations. Implementations of this :class:`.BaseEstimatorV2`
    interface must define their own :meth:`.run` method, which is designed to
    take the following inputs:

     * pubs: list of pubs (Primitive Unified Blocs). An estimator pub is a list
        or tuple of two to four elements that define the unit of work for the
        estimator. These are:

        * A single :class:`~qiskit.circuit.QuantumCircuit`, possibly parametrized,
            whose final state we define as :math:`\psi(\theta)`.

        * One or more observables (specified as any :class:`~.ObservablesArrayLike`, including
            :class:`~.Pauli`, :class:`~.SparsePauliOp`, ``str``) that specify which expectation
            values to estimate, denoted :math:`H_j`.

        * A collection parameter value sets to bind the circuit against, :math:`\theta_k`

        * Optionally, the estimation precision.

     * precision: the estimation precision. This specification is optional and will be overriden by
        the pub-wise shots if provided.

    All estimator implementations must implement default value for the ``precision`` in the
    :meth:`.run` method. This default value will be used any time ``precision=None`` is specified, which
    can take place in the :meth:`.run` kwargs or at the pub level.
    """

    @staticmethod
    def _make_data_bin(_: EstimatorPub) -> type[DataBin]:
        # this method is present for backwards compat. new primitive implementations
        # should avoid it.
        return DataBin

    @abstractmethod
    def run(
        self, pubs: Iterable[EstimatorPubLike], *, precision: float | None = None
    ) -> BasePrimitiveJob[PrimitiveResult[PubResult]]:
        """Estimate expectation values for each provided pub (Primitive Unified Bloc).

        Args:
            pubs: An iterable of pub-like objects, such as tuples ``(circuit, observables)``
                  or ``(circuit, observables, parameter_values)``.
            precision: The target precision for expectation value estimates of each
                       run Estimator Pub that does not specify its own precision. If None
                       the estimator's default precision value will be used.

        Returns:
            A job object that contains results.
        """


class BaseEstimatorV1(BasePrimitiveV1, Generic[T]):
    r"""Base class for ``EstimatorV1`` implementations.

    Note that the reference estimator in Qiskit follows the ``EstimatorV2``
    interface specifications instead.

    An estimator calculates expectation values for provided quantum circuit and
    observable combinations.

    Implementations of :class:`.BaseEstimatorV1` should define their own
    :meth:`.BaseEstimatorV1._run` method
    that will be called by the public-facing :meth:`qiskit.primitives.BaseEstimatorV1.run`,
    which takes the following inputs:

    * quantum circuits (:math:`\psi_i(\theta)`): list of (parameterized) quantum circuits
      (a list of :class:`~qiskit.circuit.QuantumCircuit` objects).

    * observables (:math:`H_j`): a list of :class:`~qiskit.quantum_info.SparsePauliOp`
      objects.

    * parameter values (:math:`\theta_k`): list of sets of values
      to be bound to the parameters of the quantum circuits
      (list of list of float).

    The method returns a :class:`~qiskit.providers.JobV1` object. Calling
    :meth:`qiskit.providers.JobV1.result()` yields the
    a list of expectation values plus optional metadata like confidence intervals for
    the estimation.

    .. math::

        \langle\psi_i(\theta_k)|H_j|\psi_i(\theta_k)\rangle

    Here is an example of how a :class:`.BaseEstimatorV1` would be used:

    .. code-block:: python

        # This is a fictional import path.
        # There are currently no EstimatorV1 implementations in Qiskit.
        from estimator_v1_location import EstimatorV1
        from qiskit.circuit.library import RealAmplitudes
        from qiskit.quantum_info import SparsePauliOp

        psi1 = RealAmplitudes(num_qubits=2, reps=2)
        psi2 = RealAmplitudes(num_qubits=2, reps=3)

        H1 = SparsePauliOp.from_list([("II", 1), ("IZ", 2), ("XI", 3)])
        H2 = SparsePauliOp.from_list([("IZ", 1)])
        H3 = SparsePauliOp.from_list([("ZI", 1), ("ZZ", 1)])

        theta1 = [0, 1, 1, 2, 3, 5]
        theta2 = [0, 1, 1, 2, 3, 5, 8, 13]
        theta3 = [1, 2, 3, 4, 5, 6]

        estimator = EstimatorV1()

        # calculate [ <psi1(theta1)|H1|psi1(theta1)> ]
        job = estimator.run([psi1], [H1], [theta1])
        job_result = job.result() # It will block until the job finishes.
        print(f"The primitive-job finished with result {job_result}")

        # calculate [ <psi1(theta1)|H1|psi1(theta1)>,
        #             <psi2(theta2)|H2|psi2(theta2)>,
        #             <psi1(theta3)|H3|psi1(theta3)> ]
        job2 = estimator.run([psi1, psi2, psi1], [H1, H2, H3], [theta1, theta2, theta3])
        job_result = job2.result()
        print(f"The primitive-job finished with result {job_result}")
    """

    __hash__ = None

    def __init__(
        self,
        *,
        options: dict | None = None,
    ):
        """
        Initialize ``EstimatorV1``.

        Args:
            options: Default options.
        """
        super().__init__(options)

    def run(
        self,
        circuits: Sequence[QuantumCircuit] | QuantumCircuit,
        observables: Sequence[BaseOperator | str] | BaseOperator | str,
        parameter_values: Sequence[Sequence[float]] | Sequence[float] | float | None = None,
        **run_options,
    ) -> T:
        """Run the job of the estimation of expectation value(s).

        ``circuits``, ``observables``, and ``parameter_values`` should have the same
        length. The i-th element of the result is the expectation of observable

        .. code-block:: python

            obs = observables[i]

        for the state prepared by

        .. code-block:: python

            circ = circuits[i]

        with bound parameters

        .. code-block:: python

            values = parameter_values[i]

        Args:
            circuits: one or more circuit objects.
            observables: one or more observable objects. Several formats are allowed;
                importantly, ``str`` should follow the string representation format for
                :class:`~qiskit.quantum_info.Pauli` objects.
            parameter_values: concrete parameters to be bound.
            run_options: runtime options used for circuit execution.

        Returns:
            The job object of EstimatorResult.

        Raises:
            TypeError: Invalid argument type given.
            ValueError: Invalid argument values given.
        """
        # Validation
        circuits, observables, parameter_values = _validate_estimator_args(
            circuits, observables, parameter_values
        )

        # Options
        run_opts = copy(self.options)
        run_opts.update_options(**run_options)

        return self._run(
            circuits,
            observables,
            parameter_values,
            **run_opts.__dict__,
        )

    @abstractmethod
    def _run(
        self,
        circuits: tuple[QuantumCircuit, ...],
        observables: tuple[SparsePauliOp, ...],
        parameter_values: tuple[tuple[float, ...], ...],
        **run_options,
    ) -> T:
        raise NotImplementedError("The subclass of BaseEstimatorV1 must implement `_run` method.")



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Compute the weighted sum of qubit states."""

from __future__ import annotations

from typing import List, Optional
import numpy as np

from qiskit.circuit import QuantumRegister, AncillaRegister, QuantumCircuit, Gate

from ..blueprintcircuit import BlueprintCircuit


class WeightedAdder(BlueprintCircuit):
    r"""A circuit to compute the weighted sum of qubit registers.

    Given :math:`n` qubit basis states :math:`q_0, \ldots, q_{n-1} \in \{0, 1\}` and non-negative
    integer weights :math:`\lambda_0, \ldots, \lambda_{n-1}`, this circuit performs the operation

    .. math::

        |q_0 \ldots q_{n-1}\rangle |0\rangle_s
        \mapsto |q_0 \ldots q_{n-1}\rangle |\sum_{j=0}^{n-1} \lambda_j q_j\rangle_s

    where :math:`s` is the number of sum qubits required.
    This can be computed as

    .. math::

        s = 1 + \left\lfloor \log_2\left( \sum_{j=0}^{n-1} \lambda_j \right) \right\rfloor

    or :math:`s = 1` if the sum of the weights is 0 (then the expression in the logarithm is
    invalid).

    For qubits in a circuit diagram, the first weight applies to the upper-most qubit.
    For an example where the state of 4 qubits is added into a sum register, the circuit can
    be schematically drawn as

    .. code-block:: text

                   ┌────────┐
          state_0: ┤0       ├ | state_0 * weights[0]
                   │        │ |
          state_1: ┤1       ├ | + state_1 * weights[1]
                   │        │ |
          state_2: ┤2       ├ | + state_2 * weights[2]
                   │        │ |
          state_3: ┤3       ├ | + state_3 * weights[3]
                   │        │
            sum_0: ┤4       ├ |
                   │  Adder │ |
            sum_1: ┤5       ├ | = sum_0 * 2^0 + sum_1 * 2^1 + sum_2 * 2^2
                   │        │ |
            sum_2: ┤6       ├ |
                   │        │
          carry_0: ┤7       ├
                   │        │
          carry_1: ┤8       ├
                   │        │
        control_0: ┤9       ├
                   └────────┘
    """

    def __init__(
        self,
        num_state_qubits: Optional[int] = None,
        weights: Optional[List[int]] = None,
        name: str = "adder",
    ) -> None:
        """Computes the weighted sum controlled by state qubits.

        Args:
            num_state_qubits: The number of state qubits.
            weights: List of weights, one for each state qubit. If none are provided they
                default to 1 for every qubit.
            name: The name of the circuit.
        """
        super().__init__(name=name)

        self._weights = None
        self._num_state_qubits = None

        self.weights = weights
        self.num_state_qubits = num_state_qubits

    @property
    def num_sum_qubits(self) -> int:
        """The number of sum qubits in the circuit.

        Returns:
            The number of qubits needed to represent the weighted sum of the qubits.
        """
        if sum(self.weights) > 0:
            return int(np.floor(np.log2(sum(self.weights))) + 1)
        return 1

    @property
    def weights(self) -> List[int]:
        """The weights for the qubit states.

        Returns:
            The weight for the qubit states.
        """
        if self._weights:
            return self._weights
        if self.num_state_qubits:
            return [1] * self.num_state_qubits
        return None

    @weights.setter
    def weights(self, weights: List[int]) -> None:
        """Set the weights for summing the qubit states.

        Args:
            weights: The new weights.

        Raises:
            ValueError: If not all weights are close to an integer.
        """
        if weights:
            for i, weight in enumerate(weights):
                if not np.isclose(weight, np.round(weight)):
                    raise ValueError("Non-integer weights are not supported!")
                weights[i] = np.round(weight)

        self._invalidate()
        self._weights = weights
        self._reset_registers()

    @property
    def num_state_qubits(self) -> int:
        """The number of qubits to be summed.

        Returns:
            The number of state qubits.
        """
        return self._num_state_qubits

    @num_state_qubits.setter
    def num_state_qubits(self, num_state_qubits: int) -> None:
        """Set the number of state qubits.

        Args:
            num_state_qubits: The new number of state qubits.
        """
        if self._num_state_qubits is None or num_state_qubits != self._num_state_qubits:
            self._invalidate()
            self._num_state_qubits = num_state_qubits
            self._reset_registers()

    def _reset_registers(self):
        """Reset the registers."""
        self.qregs = []

        if self.num_state_qubits:
            qr_state = QuantumRegister(self.num_state_qubits, name="state")
            qr_sum = QuantumRegister(self.num_sum_qubits, name="sum")
            self.qregs = [qr_state, qr_sum]

            if self.num_carry_qubits > 0:
                qr_carry = AncillaRegister(self.num_carry_qubits, name="carry")
                self.add_register(qr_carry)

            if self.num_control_qubits > 0:
                qr_control = AncillaRegister(self.num_control_qubits, name="control")
                self.add_register(qr_control)

    @property
    def num_carry_qubits(self) -> int:
        """The number of carry qubits required to compute the sum.

        Note that this is not necessarily equal to the number of ancilla qubits, these can
        be queried using ``num_ancilla_qubits``.

        Returns:
            The number of carry qubits required to compute the sum.
        """
        return self.num_sum_qubits - 1

    @property
    def num_control_qubits(self) -> int:
        """The number of additional control qubits required.

        Note that the total number of ancilla qubits can be obtained by calling the
        method ``num_ancilla_qubits``.

        Returns:
            The number of additional control qubits required (0 or 1).
        """
        return int(self.num_sum_qubits > 2)

    def _check_configuration(self, raise_on_failure=True):
        """Check if the current configuration is valid."""
        valid = True
        if self._num_state_qubits is None:
            valid = False
            if raise_on_failure:
                raise AttributeError("The number of state qubits has not been set.")

        if self._num_state_qubits != len(self.weights):
            valid = False
            if raise_on_failure:
                raise ValueError("Mismatching number of state qubits and weights.")

        return valid

    def _build(self):
        """If not already built, build the circuit."""
        if self._is_built:
            return

        super()._build()

        num_result_qubits = self.num_state_qubits + self.num_sum_qubits

        circuit = QuantumCircuit(*self.qregs)
        qr_state = circuit.qubits[: self.num_state_qubits]
        qr_sum = circuit.qubits[self.num_state_qubits : num_result_qubits]
        qr_carry = circuit.qubits[num_result_qubits : num_result_qubits + self.num_carry_qubits]
        qr_control = circuit.qubits[num_result_qubits + self.num_carry_qubits :]

        # loop over state qubits and corresponding weights
        for i, weight in enumerate(self.weights):
            # only act if non-trivial weight
            if np.isclose(weight, 0):
                continue

            # get state control qubit
            q_state = qr_state[i]

            # get bit representation of current weight
            weight_binary = f"{int(weight):b}".rjust(self.num_sum_qubits, "0")[::-1]

            # loop over bits of current weight and add them to sum and carry registers
            for j, bit in enumerate(weight_binary):
                if bit == "1":
                    if self.num_sum_qubits == 1:
                        circuit.cx(q_state, qr_sum[j])
                    elif j == 0:
                        # compute (q_sum[0] + 1) into (q_sum[0], q_carry[0])
                        # - controlled by q_state[i]
                        circuit.ccx(q_state, qr_sum[j], qr_carry[j])
                        circuit.cx(q_state, qr_sum[j])
                    elif j == self.num_sum_qubits - 1:
                        # compute (q_sum[j] + q_carry[j-1] + 1) into (q_sum[j])
                        # - controlled by q_state[i] / last qubit,
                        # no carry needed by construction
                        circuit.cx(q_state, qr_sum[j])
                        circuit.ccx(q_state, qr_carry[j - 1], qr_sum[j])
                    else:
                        # compute (q_sum[j] + q_carry[j-1] + 1) into (q_sum[j], q_carry[j])
                        # - controlled by q_state[i]
                        circuit.x(qr_sum[j])
                        circuit.x(qr_carry[j - 1])
                        circuit.mcx(
                            [q_state, qr_sum[j], qr_carry[j - 1]],
                            qr_carry[j],
                            qr_control,
                            mode="v-chain",
                        )
                        circuit.cx(q_state, qr_carry[j])
                        circuit.x(qr_sum[j])
                        circuit.x(qr_carry[j - 1])
                        circuit.cx(q_state, qr_sum[j])
                        circuit.ccx(q_state, qr_carry[j - 1], qr_sum[j])
                else:
                    if self.num_sum_qubits == 1:
                        pass  # nothing to do, since nothing to add
                    elif j == 0:
                        pass  # nothing to do, since nothing to add
                    elif j == self.num_sum_qubits - 1:
                        # compute (q_sum[j] + q_carry[j-1]) into (q_sum[j])
                        # - controlled by q_state[i] / last qubit,
                        # no carry needed by construction
                        circuit.ccx(q_state, qr_carry[j - 1], qr_sum[j])
                    else:
                        # compute (q_sum[j] + q_carry[j-1]) into (q_sum[j], q_carry[j])
                        # - controlled by q_state[i]
                        circuit.mcx(
                            [q_state, qr_sum[j], qr_carry[j - 1]],
                            qr_carry[j],
                            qr_control,
                            mode="v-chain",
                        )
                        circuit.ccx(q_state, qr_carry[j - 1], qr_sum[j])

            # uncompute carry qubits
            for j in reversed(range(len(weight_binary))):
                bit = weight_binary[j]
                if bit == "1":
                    if self.num_sum_qubits == 1:
                        pass
                    elif j == 0:
                        circuit.x(qr_sum[j])
                        circuit.ccx(q_state, qr_sum[j], qr_carry[j])
                        circuit.x(qr_sum[j])
                    elif j == self.num_sum_qubits - 1:
                        pass
                    else:
                        circuit.x(qr_carry[j - 1])
                        circuit.mcx(
                            [q_state, qr_sum[j], qr_carry[j - 1]],
                            qr_carry[j],
                            qr_control,
                            mode="v-chain",
                        )
                        circuit.cx(q_state, qr_carry[j])
                        circuit.x(qr_carry[j - 1])
                else:
                    if self.num_sum_qubits == 1:
                        pass
                    elif j == 0:
                        pass
                    elif j == self.num_sum_qubits - 1:
                        pass
                    else:
                        # compute (q_sum[j] + q_carry[j-1]) into (q_sum[j], q_carry[j])
                        # - controlled by q_state[i]
                        circuit.x(qr_sum[j])
                        circuit.mcx(
                            [q_state, qr_sum[j], qr_carry[j - 1]],
                            qr_carry[j],
                            qr_control,
                            mode="v-chain",
                        )
                        circuit.x(qr_sum[j])

        self.append(circuit.to_gate(), self.qubits)


class WeightedSumGate(Gate):
    r"""A gate to compute the weighted sum of qubit registers.

    Given :math:`n` qubit basis states :math:`q_0, \ldots, q_{n-1} \in \{0, 1\}` and non-negative
    integer weights :math:`\lambda_0, \ldots, \lambda_{n-1}`, this implements the operation

    .. math::

        |q_0 \ldots q_{n-1}\rangle |0\rangle_s
        \mapsto |q_0 \ldots q_{n-1}\rangle |\sum_{j=0}^{n-1} \lambda_j q_j\rangle_s

    where :math:`s` is the number of sum qubits required.
    This can be computed as

    .. math::

        s = 1 + \left\lfloor \log_2\left( \sum_{j=0}^{n-1} \lambda_j \right) \right\rfloor

    or :math:`s = 1` if the sum of the weights is 0 (then the expression in the logarithm is
    invalid).

    For qubits in a circuit diagram, the first weight applies to the upper-most qubit.
    For an example where the state of 4 qubits is added into a sum register, the circuit can
    be schematically drawn as

    .. code-block:: text

                   ┌──────────────┐
          state_0: ┤0             ├ | state_0 * weights[0]
                   │              │ |
          state_1: ┤1             ├ | + state_1 * weights[1]
                   │              │ |
          state_2: ┤2             ├ | + state_2 * weights[2]
                   │              │ |
          state_3: ┤3 WeightedSum ├ | + state_3 * weights[3]
                   │              │
            sum_0: ┤4             ├ |
                   │              │ |
            sum_1: ┤5             ├ | = sum_0 * 2^0 + sum_1 * 2^1 + sum_2 * 2^2
                   │              │ |
            sum_2: ┤6             ├ |
                   └──────────────┘
    """

    def __init__(
        self,
        num_state_qubits: int,
        weights: list[int] | None = None,
        label: str | None = None,
    ) -> None:
        """
        Args:
            num_state_qubits: The number of state qubits.
            weights: List of weights, one for each state qubit. If none are provided they
                default to 1 for every qubit.
            label: The name of the circuit.
        """
        if weights is None:
            weights = [1] * num_state_qubits

        self.num_state_qubits = num_state_qubits

        if sum(weights) > 0:
            self.num_sum_qubits = int(np.floor(np.log2(sum(weights))) + 1)
        else:
            self.num_sum_qubits = 1

        super().__init__("WeightedSum", self.num_state_qubits + self.num_sum_qubits, weights, label)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Synthesis for arithmetic circuits."""

from .comparators import synth_integer_comparator_2s, synth_integer_comparator_greedy
from .adders import adder_qft_d00, adder_ripple_c04, adder_ripple_v95
from .multipliers import multiplier_cumulative_h18, multiplier_qft_r17
from .weighted_sum import synth_weighted_sum_carry



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
=====================================
QPY serialization (:mod:`qiskit.qpy`)
=====================================

.. currentmodule:: qiskit.qpy

QPY is a binary serialization format for :class:`~.QuantumCircuit`
objects that is designed to be cross-platform, Python version agnostic,
and backwards compatible moving forward. QPY should be used if you need
a mechanism to save or copy between systems a :class:`~.QuantumCircuit`
that preserves the full Qiskit object structure (except for custom attributes
defined outside of Qiskit code). This differs from other serialization formats like
`OpenQASM <https://github.com/openqasm/openqasm>`__ (2.0 or 3.0) which has a
different abstraction model and can result in a loss of information contained
in the original circuit (or is unable to represent some aspects of the
Qiskit objects) or Python's `pickle <https://docs.python.org/3/library/pickle.html>`__
which will preserve the Qiskit object exactly but will only work for a single Qiskit
version (it is also
`potentially insecure <https://docs.python.org/3/library/pickle.html#module-pickle>`__).

Basic Usage
===========

Using QPY is defined to be straightforward and mirror the user API of the
serializers in Python's standard library, ``pickle`` and ``json``. There are
2 user facing functions: :func:`qiskit.qpy.dump` and
:func:`qiskit.qpy.load` which are used to dump QPY data
to a file object and load circuits from QPY data in a file object respectively.
For example:

.. plot::
    :nofigs:
    :context: reset

    # This code is hidden from users
    # It's a hack to avoid writing to file when testing the code examples
    import io
    bytestream = io.BytesIO()
    bytestream.close = lambda: bytestream.seek(0)
    def open(*args):
        return bytestream

.. plot::
    :include-source:
    :nofigs:
    :context:

    from qiskit.circuit import QuantumCircuit
    from qiskit import qpy

    qc = QuantumCircuit(2, name='Bell', metadata={'test': True})
    qc.h(0)
    qc.cx(0, 1)
    qc.measure_all()

    with open('bell.qpy', 'wb') as fd:
        qpy.dump(qc, fd)

    with open('bell.qpy', 'rb') as fd:
        new_qc = qpy.load(fd)[0]

The :func:`qiskit.qpy.dump` function also lets you
include multiple circuits in a single QPY file:

.. plot::
    :include-source:
    :nofigs:
    :context:

    with open('twenty_bells.qpy', 'wb') as fd:
        qpy.dump([qc] * 20, fd)

and then loading that file will return a list with all the circuits

.. plot::
    :include-source:
    :nofigs:
    :context:

    with open('twenty_bells.qpy', 'rb') as fd:
        twenty_new_bells = qpy.load(fd)


API documentation
=================

.. autofunction:: load
.. autofunction:: dump

These functions will raise a custom subclass of :exc:`.QiskitError` if they encounter problems
during serialization or deserialization.

.. autoexception:: QpyError

When a lower-than-maximum target QPY version is set for serialization, but the object to be
serialized contains features that cannot be represented in that format, a subclass of
:exc:`QpyError` is raised:

.. autoexception:: UnsupportedFeatureForVersion

Attributes:
    QPY_VERSION (int): The current QPY format version as of this release. This
        is the default value of the ``version`` keyword argument on
        :func:`.qpy.dump` and also the upper bound for accepted values for
        the same argument. This is also the upper bond on the versions supported
        by :func:`.qpy.load`.

    QPY_COMPATIBILITY_VERSION (int): The current minimum compatibility QPY
        format version. This is the minimum version that :func:`.qpy.dump`
        will accept for the ``version`` keyword argument. :func:`.qpy.load`
        will be able to load all released format versions of QPY (up until
        ``QPY_VERSION``).

.. _qpy_compatibility:

QPY Compatibility
=================

The QPY format is designed to be backwards compatible moving forward. This means
you should be able to load a QPY with any newer Qiskit version than the one
that generated it. However, loading a QPY file with an older Qiskit version is
not supported and may not work.

For example, if you generated a QPY file using qiskit-terra 0.18.1 you could
load that QPY file with qiskit-terra 0.19.0 and a hypothetical qiskit-terra
0.29.0. However, loading that QPY file with 0.18.0 is not supported and may not
work.

If a feature being loaded is deprecated in the corresponding qiskit release, QPY will
raise a :exc:`~.QPYLoadingDeprecatedFeatureWarning` informing of the deprecation period
and how the feature will be internally handled.

.. autoexception:: QPYLoadingDeprecatedFeatureWarning

.. note::

    With versions of Qiskit before 1.2.4, the ``use_symengine=True`` argument to :func:`.qpy.dump`
    could cause problems with backwards compatibility if there were :class:`.ParameterExpression`
    objects to serialize.  In particular:

    * When the loading version of Qiskit is 1.2.4 or greater, QPY files generated with any version
      of Qiskit >= 0.46.0 can be loaded.  If a version of Qiskit between 0.45.0 and 0.45.3 was used
      to generate the files, and the non-default argument ``use_symengine=True`` was given to
      :func:`.qpy.dump`, the file can only be read if the version of ``symengine`` used in the
      generating environment was in the 0.11 or 0.13 series, but if the environment was created
      during the support window of Qiskit 0.45, it is likely that ``symengine==0.9.2`` was used.

    * When the loading version of Qiskit is between 0.46.0 and 1.2.2 inclusive, the file can only be
      read if the installed version of ``symengine`` in the loading environment matches the version
      used in the generating environment.

    To recover a QPY file that fails with ``symengine`` version-related errors during a call to
    :func:`.qpy.load`, first attempt to use Qiskit >= 1.2.4 to load the file.  If this still fails,
    it is likely because Qiskit 0.45.x was used to generate the file with ``use_symengine=True``.
    In this case, use Qiskit 0.45.3 with ``symengine==0.9.2`` to load the file, and then re-export
    it to QPY setting ``use_symengine=False``.  The resulting file can then be loaded by any later
    version of Qiskit.

.. note::

    Starting with Qiskit version 2.0.0, which removed the Pulse module from the library, QPY provides
    limited support for loading payloads that include pulse data. Loading a ``ScheduleBlock`` payload,
    a :class:`.QpyError` exception will be raised. Loading a payload for a circuit that contained pulse
    gates, the output circuit will contain  custom instructions **without** calibration data attached
    for each pulse gate, leaving them undefined.

QPY format version history
--------------------------

If you're planning to load a QPY file between different Qiskit versions knowing
which versions were available in a given release are useful. As the QPY is
backwards compatible but not forwards compatible you need to ensure a given
QPY format version was released in the release you're calling :func:`.load`
with. The following table lists the QPY versions that were supported in every
Qiskit (and qiskit-terra prior to Qiskit 1.0.0) release going back to the introduction
of QPY in qiskit-terra 0.18.0.

.. list-table:: QPY Format Version History
   :header-rows: 1

   * - Qiskit (qiskit-terra for < 1.0.0) version
     - :func:`.dump` format(s) output versions
     - :func:`.load` maximum supported version (older format versions can always be read)
   * - 1.3.0
     - 10, 11, 12, 13
     - 13
   * - 1.2.4
     - 10, 11, 12
     - 12
   * - 1.2.3 (yanked)
     - 10, 11, 12
     - 12
   * - 1.2.2
     - 10, 11, 12
     - 12
   * - 1.2.1
     - 10, 11, 12
     - 12
   * - 1.2.0
     - 10, 11, 12
     - 12
   * - 1.1.0
     - 10, 11, 12
     - 12
   * - 1.0.2
     - 10, 11
     - 11
   * - 1.0.1
     - 10, 11
     - 11
   * - 1.0.0
     - 10, 11
     - 11
   * - 0.46.1
     - 10
     - 10
   * - 0.45.3
     - 10
     - 10
   * - 0.45.2
     - 10
     - 10
   * - 0.45.1
     - 10
     - 10
   * - 0.45.0
     - 10
     - 10
   * - 0.25.3
     - 9
     - 9
   * - 0.25.2
     - 9
     - 9
   * - 0.25.1
     - 9
     - 9
   * - 0.24.2
     - 8
     - 8
   * - 0.24.1
     - 7
     - 7
   * - 0.24.0
     - 7
     - 7
   * - 0.23.3
     - 6
     - 6
   * - 0.23.2
     - 6
     - 6
   * - 0.23.1
     - 6
     - 6
   * - 0.23.0
     - 6
     - 6
   * - 0.22.4
     - 5
     - 5
   * - 0.22.3
     - 5
     - 5
   * - 0.22.2
     - 5
     - 5
   * - 0.22.1
     - 5
     - 5
   * - 0.22.0
     - 5
     - 5
   * - 0.21.2
     - 5
     - 5
   * - 0.21.1
     - 5
     - 5
   * - 0.21.0
     - 5
     - 5
   * - 0.20.2
     - 4
     - 4
   * - 0.20.1
     - 4
     - 4
   * - 0.20.0
     - 4
     - 4
   * - 0.19.2
     - 4
     - 4
   * - 0.19.1
     - 3
     - 3
   * - 0.19.0
     - 2
     - 2
   * - 0.18.3
     - 1
     - 1
   * - 0.18.2
     - 1
     - 1
   * - 0.18.1
     - 1
     - 1
   * - 0.18.0
     - 1
     - 1

.. _qpy_format:

QPY Format
==========

The QPY serialization format is a portable cross-platform binary
serialization format for :class:`~qiskit.circuit.QuantumCircuit` objects in Qiskit. The basic
file format is as follows:

A QPY file (or memory object) always starts with the following 6
byte UTF8 string: ``QISKIT`` which is immediately followed by the overall
file header. The contents of the file header as defined as a C struct are:

.. code-block:: c

    struct {
        uint8_t qpy_version;
        uint8_t qiskit_major_version;
        uint8_t qiskit_minor_version;
        uint8_t qiskit_patch_version;
        uint64_t num_circuits;
    }


From V10 on, a new field is added to the file header struct to represent the
encoding scheme used for symbolic expressions:

.. code-block:: c

    struct {
        uint8_t qpy_version;
        uint8_t qiskit_major_version;
        uint8_t qiskit_minor_version;
        uint8_t qiskit_patch_version;
        uint64_t num_circuits;
        char symbolic_encoding;
    }

All values use network byte order [#f1]_ (big endian) for cross platform
compatibility.

The file header is immediately followed by the circuit payloads.
Each individual circuit is composed of the following parts:

``HEADER | METADATA | REGISTERS | STANDALONE_VARS | CUSTOM_DEFINITIONS | INSTRUCTIONS``

The ``STANDALONE_VARS`` are new in QPY version 12; before that, there was no data between
``REGISTERS`` and ``CUSTOM_DEFINITIONS``.

There is a circuit payload for each circuit (where the total number is dictated
by ``num_circuits`` in the file header). There is no padding between the
circuits in the data.

.. _qpy_version_14:

Version 14
----------

Version 14 adds a new core DURATION type, support for additional :class:`~.types.Type`
classes :class:`~.types.Float` and :class:`~.types.Duration`, and a new expression
node type :class:`~.expr.Stretch`.

DURATION
~~~~~~~~

A :class:`~.circuit.Duration` is encoded by a single-byte ASCII ``char`` that encodes the kind of
type, followed by a payload that varies depending on the type.  The defined codes are:

==============================  =========  =========================================================
Qiskit class                    Type code  Payload
==============================  =========  =========================================================
:class:`~.circuit.Duration.dt`   ``t``     One ``unsigned long long value``.

:class:`~.circuit.Duration.ns`   ``n``     One ``double value``.

:class:`~.circuit.Duration.us`   ``u``     One ``double value``.

:class:`~.circuit.Duration.ms`   ``m``     One ``double value``.

:class:`~.circuit.Duration.s`    ``s``     One ``double value``.

==============================  =========  =========================================================

Changes to EXPR_VAR_DECLARATION
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The ``EXPR_VAR_DECLARATION`` type is now used to represent both :class:`~.expr.Var` standalone
variables and :class:`~.expr.Stretch` identifiers. To support this change, the usage type code has
two new possible entries, in addition to the existing ones:

=========  =========================================================================================
Type code  Meaning
=========  =========================================================================================
``A``      A ``capture`` stretch to the circuit.

``O``      A locally declared stretch to the circuit.

=========  =========================================================================================

Changes to EXPRESSION
---------------------

The EXPRESSION type code has a new possible entry, ``s``, corresponding to :class:`.expr.Stretch`
nodes.

=======================  =========  ======================================================  ========
Qiskit class             Type code  Payload                                                 Children
=======================  =========  ======================================================  ========
:class:`~.expr.Stretch`  ``s``      One ``unsigned short var_index``                        0
=======================  =========  ======================================================  ========

Changes to EXPR_TYPE
~~~~~~~~~~~~~~~~~~~~

The following table shows the new type classes added in the version:

=========================  =========  ==============================================================
Qiskit class               Type code  Payload
=========================  =========  ==============================================================
:class:`~.types.Float`     ``f``      None.

:class:`~.types.Duration`  ``d``      None.

=========================  =========  ==============================================================

Changes to EXPR_VALUE
~~~~~~~~~~~~~~~~~~~~~

The classical expression's type system now supports new encoding types for value literals, in
addition to the existing encodings for int and bool. The new value type encodings are below:

===========================  =========  ============================================================
Python type                  Type code  Payload
===========================  =========  ============================================================
``float``                    ``f``      One ``double value``.

:class:`~.circuit.Duration`  ``t``      One ``DURATION``.

===========================  =========  ============================================================

.. _qpy_version_13:

Version 13
----------

Version 13 added a native Qiskit serialization representation for :class:`.ParameterExpression`.
Previous QPY versions relied on either ``sympy`` or ``symengine`` to serialize the underlying symbolic
expression. Starting in Version 13, QPY now represents the sequence of API calls used to create the
:class:`.ParameterExpression`.

The main change in the serialization format is in the :ref:`qpy_param_expr_v3` payload.  The
``expr_size`` bytes following the head now contain an array of ``PARAM_EXPR_ELEM_V13`` structs. The
intent is for this array to be read one struct at a time, where each struct describes one of the
calls to make to reconstruct the :class:`.ParameterExpression`.

PARAM_EXPR_ELEM_V13
~~~~~~~~~~~~~~~~~~~

The struct format is defined as:

.. code-block:: c

    struct {
        unsigned char op_code;
        char lhs_type;
        char lhs[16];
        char rhs_type;
        char rhs[16];
    } PARAM_EXPR_ELEM_V13;

The ``op_code`` field is used to define the operation added to the :class:`.ParameterExpression`.
The value can be:

.. list-table:: PARAM_EXPR_ELEM_V13 op code values
   :header-rows: 1

   * - ``op_code``
     - :class:`.ParameterExpression` method
   * - 0
     - :meth:`~.ParameterExpression.__add__`
   * - 1
     - :meth:`~.ParameterExpression.__sub__`
   * - 2
     - :meth:`~.ParameterExpression.__mul__`
   * - 3
     - :meth:`~.ParameterExpression.__truediv__`
   * - 4
     - :meth:`~.ParameterExpression.__pow__`
   * - 5
     - :meth:`~.ParameterExpression.sin`
   * - 6
     - :meth:`~.ParameterExpression.cos`
   * - 7
     - :meth:`~.ParameterExpression.tan`
   * - 8
     - :meth:`~.ParameterExpression.arcsin`
   * - 9
     - :meth:`~.ParameterExpression.arccos`
   * - 10
     - :meth:`~.ParameterExpression.exp`
   * - 11
     - :meth:`~.ParameterExpression.log`
   * - 12
     - :meth:`~.ParameterExpression.sign`
   * - 13
     - :meth:`~.ParameterExpression.gradient`
   * - 14
     - :meth:`~.ParameterExpression.conjugate`
   * - 15
     - :meth:`~.ParameterExpression.subs`
   * - 16
     - :meth:`~.ParameterExpression.abs`
   * - 17
     - :meth:`~.ParameterExpression.arctan`
   * - 255
     - NULL

The ``NULL`` value of 255 is only used to fill the op code field for
entries that are not actual operations but indicate recursive definitions.
Then the ``lhs_type`` and ``rhs_type`` fields are used to describe
the operand types and can be one of the following UTF-8 encoded
characters:

.. list-table:: PARAM_EXPR_ELEM_V13 operand type values
   :header-rows: 1

   * - Value
     - Type
   * - ``n``
     - ``None``
   * - ``p``
     - :class:`.Parameter`
   * - ``f``
     - ``float``
   * - ``c``
     - ``complex``
   * - ``i``
     - ``int``
   * - ``s``
     - Recursive :class:`.ParameterExpression` definition start
   * - ``e``
     - Recursive :class:`.ParameterExpression` definition stop
   * - ``u``
     - substitution

If the type value is ``f``, ``c``, or ``i``, the corresponding ``lhs`` or ``rhs``
field widths are 128 bits each. In the case of floats, the literal value is encoded as a double
with 0 padding, while complex numbers are encoded as real part followed by imaginary part,
taking up 64 bits each. For ``i``, the value is encoded as a 64 bit signed integer with 0 padding
for the full 128 bit width. ``n`` is used to represent a ``None`` and typically isn't directly used
as it indicates an argument that's not used. For ``p`` the data is the UUID for the
:class:`.Parameter` which can be looked up in the symbol map described in the
``map_elements`` outer :ref:`qpy_param_expr_v3` payload. If the type value is
``s`` this marks the start of a a new recursive section for a nested
:class:`.ParameterExpression`. For example, in the following snippet there is an inner ``expr``
contained in ``final_expr``, constituting a nested expression::

    from qiskit.circuit import Parameter

    x = Parameter("x")
    y = Parameter("y")
    z = Parameter("z")

    expr = (x + y) / 2
    final_expr = z**2 + expr

When ``s`` is encountered, this indicates that until an ``e` struct is reached, the next structs
are used for a recursive definition. For both
``s`` and ``e`` types, the data values are not used, and always set to 0. The type value
of ``u`` is used to represent a substitution call. This is only used for ``lhs_type``
and is always paired with an ``rhs_type`` of ``n``. The data value is the size in bytes of
a :ref:`qpy_mapping` encoded mapping of :class:`.Parameter` names to their value for the
:meth:`~.ParameterExpression.subs` call. The mapping data is immediately following the
struct, and the next struct starts immediately after the mapping data.

.. _qpy_version_12:

Version 12
----------

Version 12 adds support for:

* circuits containing memory-owning :class:`.expr.Var` variables.

Changes to HEADER
~~~~~~~~~~~~~~~~~

The HEADER struct for an individual circuit has added three ``uint32_t`` counts of the input,
captured and locally declared variables in the circuit.  The new form looks like:

.. code-block:: c

    struct {
        uint16_t name_size;
        char global_phase_type;
        uint16_t global_phase_size;
        uint32_t num_qubits;
        uint32_t num_clbits;
        uint64_t metadata_size;
        uint32_t num_registers;
        uint64_t num_instructions;
        uint32_t num_vars;
    } HEADER_V12;

The ``HEADER_V12`` struct is followed immediately by the same name, global-phase, metadata
and register information as the V2 version of the header.  Immediately following the registers is
``num_vars`` instances of ``EXPR_VAR_STANDALONE`` that define the variables in this circuit.  After
that, the data continues with custom definitions and instructions as in prior versions of QPY.


EXPR_VAR_DECLARATION
~~~~~~~~~~~~~~~~~~~~

An ``EXPR_VAR_DECLARATION`` defines an :class:`.expr.Var` instance that is standalone; that is, it
represents a self-owned memory location rather than wrapping a :class:`.Clbit` or
:class:`.ClassicalRegister`.  The payload is a C struct:

.. code-block:: c

    struct {
        char uuid_bytes[16];
        char usage;
        uint16_t name_size;
    }

which is immediately followed by an ``EXPR_TYPE`` payload and then ``name_size`` bytes of UTF-8
encoding string data containing the name of the variable.

The ``char`` usage type code takes the following values:

=========  =========================================================================================
Type code  Meaning
=========  =========================================================================================
``I``      An ``input`` variable to the circuit.

``C``      A ``capture`` variable to the circuit.

``L``      A locally declared variable to the circuit.
=========  =========================================================================================


Changes to EXPR_VAR
~~~~~~~~~~~~~~~~~~~

The ``EXPR_VAR`` variable has gained a new type code and payload, in addition to the pre-existing ones:

===========================  =========  ============================================================
Python class                 Type code  Payload
===========================  =========  ============================================================
:class:`.UUID`               ``U``      One ``uint32_t`` index of the variable into the series of
                                        ``EXPR_VAR_STANDALONE`` variables that were written
                                        immediately after the circuit header.
===========================  =========  ============================================================

Notably, this new type-code indexes into pre-defined variables from the circuit header, rather than
redefining the variable again in each location it is used.


Changes to EXPRESSION
---------------------

The EXPRESSION type code has a new possible entry, ``i``, corresponding to :class:`.expr.Index`
nodes.

======================  =========  =======================================================  ========
Qiskit class            Type code  Payload                                                  Children
======================  =========  =======================================================  ========
:class:`~.expr.Index`   ``i``      No additional payload. The children are the target       2
                                   and the index, in that order.
======================  =========  =======================================================  ========


.. _qpy_version_11:

Version 11
----------

Version 11 is identical to Version 10 except for the following.
First, the names in the CUSTOM_INSTRUCTION blocks
have a suffix of the form ``"_{uuid_hex}"`` where ``uuid_hex`` is a uuid
hexadecimal string such as returned by :attr:`.UUID.hex`. For example:
``"b3ecab5b4d6a4eb6bc2b2dbf18d83e1e"``.
Second, it adds support for :class:`.AnnotatedOperation`
objects. The base operation of an annotated operation is stored using the INSTRUCTION block,
and an additional ``type`` value ``'a'``is added to indicate that the custom instruction is an
annotated operation. The list of modifiers are stored as instruction parameters using INSTRUCTION_PARAM,
with an additional value ``'m'`` is added to indicate that the parameter is of type
:class:`~qiskit.circuit.annotated_operation.Modifier`. Each modifier is stored using the
MODIFIER struct.

.. _modifier_qpy:

MODIFIER
~~~~~~~~

This represents :class:`~qiskit.circuit.annotated_operation.Modifier`

.. code-block:: c

    struct {
        char type;
        uint32_t num_ctrl_qubits;
        uint32_t ctrl_state;
        double power;
    }

This is sufficient to store different types of modifiers required for serializing objects
of type :class:`.AnnotatedOperation`.
The field ``type`` is either ``'i'``, ``'c'`` or ``'p'``, representing whether the modifier
is respectively an inverse modifier, a control modifier or a power modifier. In the second
case, the fields ``num_ctrl_qubits`` and ``ctrl_state`` specify the control logic of the base
operation, and in the third case the field ``power`` represents the power of the base operation.

.. _qpy_version_10:

Version 10
----------

Version 10 adds support for:

* symengine-native serialization for objects of type :class:`~.ParameterExpression` as well as
  symbolic expressions in Pulse schedule blocks.
* new fields in the :class:`~.TranspileLayout` class added in the Qiskit 0.45.0 release.

The symbolic_encoding field is added to the file header, and a new encoding type char
is introduced, mapped to each symbolic library as follows: ``p`` refers to sympy
encoding and ``e`` refers to symengine encoding.

Changes to FILE_HEADER
~~~~~~~~~~~~~~~~~~~~~~

The contents of FILE_HEADER after V10 are defined as a C struct as:

.. code-block:: c

    struct {
        uint8_t qpy_version;
        uint8_t qiskit_major_version;
        uint8_t qiskit_minor_version;
        uint8_t qiskit_patch_version;
        uint64_t num_circuits;
        char symbolic_encoding;
    } FILE_HEADER_V10;

Changes to LAYOUT
~~~~~~~~~~~~~~~~~

The ``LAYOUT`` struct is updated to have an additional ``input_qubit_count`` field.
With version 10 the ``LAYOUT`` struct is now:

.. code-block:: c

    struct {
        char exists;
        int32_t initial_layout_size;
        int32_t input_mapping_size;
        int32_t final_layout_size;
        uint32_t extra_registers;
        int32_t input_qubit_count;
    }

The rest of the layout data after the ``LAYOUT`` struct is represented as in previous versions. If
``input qubit_count`` is < 0 that indicates that both ``_input_qubit_count``
and ``_output_qubit_list`` in the :class:`~.TranspileLayout` object are ``None``.

.. _qpy_version_9:

Version 9
---------

Version 9 adds support for classical :class:`~.expr.Expr` nodes and their associated
:class:`~.types.Type`\\ s.


EXPRESSION
~~~~~~~~~~

An :class:`~.expr.Expr` node is represented by a stream of variable-width data.  A node itself is
represented by (in order in the byte stream):

#. a one-byte type code discriminator;
#. an EXPR_TYPE object;
#. a type-code-specific additional payload;
#. a type-code-specific number of child EXPRESSION payloads (the number of these is implied by the
   type code and not explicitly stored).

Each of these are described in the following table:

======================  =========  =======================================================  ========
Qiskit class            Type code  Payload                                                  Children
======================  =========  =======================================================  ========
:class:`~.expr.Var`     ``x``      One ``EXPR_VAR``.                                        0

:class:`~.expr.Value`   ``v``      One ``EXPR_VALUE``.                                      0

:class:`~.expr.Cast`    ``c``      One ``_Bool``  that corresponds to the value of          1
                                   ``implicit``.

:class:`~.expr.Unary`   ``u``      One ``uint8_t`` with the same numeric value as the       1
                                   :class:`.Unary.Op`.

:class:`~.expr.Binary`  ``b``      One ``uint8_t`` with the same numeric value as the       2
                                   :class:`.Binary.Op`.
======================  =========  =======================================================  ========


EXPR_TYPE
~~~~~~~~~

A :class:`~.types.Type` is encoded by a single-byte ASCII ``char`` that encodes the kind of type,
followed by a payload that varies depending on the type.  The defined codes are:

======================  =========  =================================================================
Qiskit class            Type code  Payload
======================  =========  =================================================================
:class:`~.types.Bool`   ``b``      None.

:class:`~.types.Uint`   ``u``      One ``uint32_t width``.
======================  =========  =================================================================


EXPR_VAR
~~~~~~~~

This represents a runtime variable of a :class:`~.expr.Var` node.  These are a type code, followed
by a type-code-specific payload:

===========================  =========  ============================================================
Python class                 Type code  Payload
===========================  =========  ============================================================
:class:`.Clbit`              ``C``      One ``uint32_t index`` that is the index of the
                                        :class:`.Clbit` in the containing circuit.

:class:`.ClassicalRegister`  ``R``      One ``uint16_t reg_name_size``, followed by that many bytes
                                        of UTF-8 string data of the register name.
===========================  =========  ============================================================


EXPR_VALUE
~~~~~~~~~~

This represents a literal object in the classical type system, such as an integer.  Currently there
are very few such literals.  These are encoded as a type code, followed by a type-code-specific
payload.

===========  =========  ============================================================================
Python type  Type code  Payload
===========  =========  ============================================================================
``bool``     ``b``      One ``_Bool value``.

``int``      ``i``      One ``uint8_t num_bytes``, followed by the integer encoded into that many
                        many bytes (network order) in a two's complement representation.
===========  =========  ============================================================================



Changes to INSTRUCTION
~~~~~~~~~~~~~~~~~~~~~~

To support the use of :class:`~.expr.Expr` nodes in the fields :attr:`.IfElseOp.condition`,
:attr:`.WhileLoopOp.condition` and :attr:`.SwitchCaseOp.target`, the INSTRUCTION struct is changed
in an ABI compatible-manner to :ref:`its previous definition <qpy_instruction_v5>`.  The new struct
is the C struct:

.. code-block:: c

    struct {
        uint16_t name_size;
        uint16_t label_size;
        uint16_t num_parameters;
        uint32_t num_qargs;
        uint32_t num_cargs;
        uint8_t conditional_key;
        uint16_t conditional_reg_name_size;
        int64_t conditional_value;
        uint32_t num_ctrl_qubits;
        uint32_t ctrl_state;
    }

where the only change is that a ``uint8_t conditional_key`` entry has replaced ``_Bool
has_conditional``.  This new ``conditional_key`` takes the following numeric values, with these
effects:

=====  =============================================================================================
Value  Effects
=====  =============================================================================================
0      The instruction has its ``.condition`` field set to ``None``.  The
       ``conditional_reg_name_size`` and ``conditional_value`` fields should be ignored.

1      The instruction has its ``.condition`` field set to a 2-tuple of either a :class:`.Clbit`
       or a :class:`.ClassicalRegister`, and a integer of value ``conditional_value``.  The
       INSTRUCTION payload, including its trailing data is parsed exactly as it would be in QPY
       versions less than 8.

2      The instruction has its ``.condition`` field set to a :class:`~.expr.Expr` node.  The
       ``conditional_reg_name_size`` and ``conditional_value`` fields should be ignored.  The data
       following the struct is followed (as in QPY versions less than 8) by ``name_size`` bytes of
       UTF-8 string data for the class name and ``label_size`` bytes of UTF-8 string data for the
       label (if any). Then, there is one INSTRUCTION_PARAM, which will contain an EXPRESSION. After
       that, parsing continues with the INSTRUCTION_ARG structs, as in previous versions of QPY.
=====  =============================================================================================


Changes to INSTRUCTION_PARAM
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A new type code ``x`` is added that defines an EXPRESSION parameter.


.. _qpy_version_8:

Version 8
---------

Version 8 adds support for handling a :class:`~.TranspileLayout` stored in the
:attr:`.QuantumCircuit.layout` attribute. In version 8 immediately following the
calibrations block at the end of the circuit payload there is now the
``LAYOUT`` struct. This struct outlines the size of the three attributes of a
:class:`~.TranspileLayout` class.

LAYOUT
~~~~~~

.. code-block:: c

    struct {
        char exists;
        int32_t initial_layout_size;
        int32_t input_mapping_size;
        int32_t final_layout_size;
        uint32_t extra_registers;
    }

If any of the signed values are ``-1`` this indicates the corresponding
attribute is ``None``.

Immediately following the ``LAYOUT`` struct there is a :ref:`qpy_registers` struct
for ``extra_registers`` (specifically the format introduced in :ref:`qpy_version_4`)
standalone register definitions that aren't present in the circuit. Then there
are ``initial_layout_size`` ``INITIAL_LAYOUT_BIT`` structs to define the
:attr:`.TranspileLayout.initial_layout` attribute.

INITIAL_LAYOUT_BIT
~~~~~~~~~~~~~~~~~~

.. code-block:: c

    struct {
        int32_t index;
        int32_t register_size;
    }

Where a value of ``-1`` indicates ``None`` (as in no register is associated
with the bit). Following each ``INITIAL_LAYOUT_BIT`` struct is ``register_size``
bytes for a ``utf8`` encoded string for the register name.

Following the initial layout there is ``input_mapping_size`` array of
``uint32_t`` integers representing the positions of the physical bit from the
initial layout. This enables constructing a list of virtual bits where the
array index is its input mapping position.

Finally, there is an array of ``final_layout_size`` ``uint32_t`` integers. Each
element is an index in the circuit's ``qubits`` attribute which enables building
a mapping from qubit starting position to the output position at the end of the
circuit.

.. _qpy_version_7:

Version 7
---------

Version 7 adds support for :class:`.~Reference` instruction and serialization of
a ``ScheduleBlock`` program while keeping its reference to subroutines::

    from qiskit import pulse
    from qiskit import qpy

    with pulse.build() as schedule:
        pulse.reference("cr45p", "q0", "q1")
        pulse.reference("x", "q0")
        pulse.reference("cr45p", "q0", "q1")

    with open('template_ecr.qpy', 'wb') as fd:
        qpy.dump(schedule, fd)

The conventional :ref:`qpy_schedule_block` data model is preserved, but in
version 7 it is immediately followed by an extra :ref:`qpy_mapping` utf8 bytes block
representing the data of the referenced subroutines.

New type key character is added to the :ref:`qpy_schedule_instructions` group
for the :class:`.~Reference` instruction.

- ``y``: :class:`~qiskit.pulse.instructions.Reference` instruction

New type key character is added to the :ref:`qpy_schedule_operands` group
for the operands of :class:`.~Reference` instruction,
which is a tuple of strings, e.g. ("cr45p", "q0", "q1").

- ``o``: string (operand string)

Note that this is the same encoding with the built-in Python string, however,
the standard value encoding in QPY uses ``s`` type character for string data,
which conflicts with the :class:`~qiskit.pulse.library.SymbolicPulse` in the scope of
pulse instruction operands. A special type character ``o`` is reserved for
the string data that appears in the pulse instruction operands.

In addition, version 7 adds two new type keys to the INSTRUCTION_PARM struct.  ``"d"`` is followed
by no data and represents the literal value :data:`.CASE_DEFAULT` for switch-statement support.
``"R"`` represents a :class:`.ClassicalRegister` or :class:`.Clbit`, and is followed by the same
format as the description of register or classical bit as used in the first element of :ref:`the
condition of an INSTRUCTION field <qpy_instructions>`.

.. _qpy_version_6:

Version 6
---------

Version 6 adds support for :class:`.~ScalableSymbolicPulse`. These objects are saved and read
like `SymbolicPulse` objects, and the class name is added to the data to correctly handle
the class selection.

`SymbolicPulse` block now starts with SYMBOLIC_PULSE_V2 header:

.. code-block:: c

    struct {
        uint16_t class_name_size;
        uint16_t type_size;
        uint16_t envelope_size;
        uint16_t constraints_size;
        uint16_t valid_amp_conditions_size;
        _bool amp_limited;
    }

The only change compared to :ref:`qpy_version_5` is the addition of `class_name_size`. The header
is then immediately followed by ``class_name_size`` utf8 bytes with the name of the class. Currently,
either `SymbolicPulse` or `ScalableSymbolicPulse` are supported. The rest of the data is then
identical to :ref:`qpy_version_5`.

.. _qpy_version_5:

Version 5
---------

Version 5 changes from :ref:`qpy_version_4` by adding support for ``ScheduleBlock``
and changing two payloads the INSTRUCTION metadata payload and the CUSTOM_INSTRUCTION block.
These now have new fields to better account for :class:`~.ControlledGate` objects in a circuit.
In addition, new payload MAP_ITEM is defined to implement the :ref:`qpy_mapping` block.

With the support of ``ScheduleBlock``, now :class:`~.QuantumCircuit` can be
serialized together with :attr:`~.QuantumCircuit.calibrations`, or
`Pulse Gates <https://docs.quantum.ibm.com/guides/pulse>`_.
In QPY version 5 and above, :ref:`qpy_circuit_calibrations` payload is
packed after the :ref:`qpy_instructions` block.

In QPY version 5 and above,

.. code-block:: c

    struct {
        char type;
    }

immediately follows the file header block to represent the program type stored in the file.

- When ``type==c``, :class:`~.QuantumCircuit` payload follows
- When ``type==s``, ``ScheduleBlock`` payload follows

.. note::

    Different programs cannot be packed together in the same file.
    You must create different files for different program types.
    Multiple objects with the same type can be saved in a single file.

.. _qpy_schedule_block:

SCHEDULE_BLOCK
~~~~~~~~~~~~~~

``ScheduleBlock`` is first supported in QPY Version 5. This allows
users to save pulse programs in the QPY binary format as follows:

.. code-block:: python

    from qiskit import pulse, qpy

    with pulse.build() as schedule:
        pulse.play(pulse.Gaussian(160, 0.1, 40), pulse.DriveChannel(0))

    with open('schedule.qpy', 'wb') as fd:
        qpy.dump(schedule, fd)

    with open('schedule.qpy', 'rb') as fd:
        new_schedule = qpy.load(fd)[0]

Note that circuit and schedule block are serialized and deserialized through
the same QPY interface. Input data type is implicitly analyzed and
no extra option is required to save the schedule block.

.. _qpy_schedule_block_header:

SCHEDULE_BLOCK_HEADER
~~~~~~~~~~~~~~~~~~~~~

``ScheduleBlock`` block starts with the following header:

.. code-block:: c

    struct {
        uint16_t name_size;
        uint64_t metadata_size;
        uint16_t num_element;
    }

which is immediately followed by ``name_size`` utf8 bytes of schedule name and
``metadata_size`` utf8 bytes of the JSON serialized metadata dictionary
attached to the schedule.

.. _qpy_schedule_alignments:

SCHEDULE_BLOCK_ALIGNMENTS
~~~~~~~~~~~~~~~~~~~~~~~~~

Then, alignment context of the schedule block starts with ``char``
representing the supported context type followed by the :ref:`qpy_sequence` block representing
the parameters associated with the alignment context :attr:`AlignmentKind._context_params`.
The context type char is mapped to each alignment subclass as follows:

- ``l``: :class:`~.AlignLeft`
- ``r``: :class:`~.AlignRight`
- ``s``: :class:`~.AlignSequential`
- ``e``: :class:`~.AlignEquispaced`

Note that :class:`~.AlignFunc` context is not supported because of the callback function
stored in the context parameters.

.. _qpy_schedule_instructions:

SCHEDULE_BLOCK_INSTRUCTIONS
~~~~~~~~~~~~~~~~~~~~~~~~~~~

This alignment block is further followed by ``num_element`` length of block elements which may
consist of nested schedule blocks and schedule instructions.
Each schedule instruction starts with ``char`` representing the instruction type
followed by the :ref:`qpy_sequence` block representing the instruction
:attr:`~qiskit.pulse.instructions.Instruction.operands`.
Note that the data structure of pulse :class:`~qiskit.pulse.instructions.Instruction`
is unified so that instance can be uniquely determined by the class and a tuple of operands.
The mapping of type char to the instruction subclass is defined as follows:

- ``a``: :class:`~qiskit.pulse.instructions.Acquire` instruction
- ``p``: :class:`~qiskit.pulse.instructions.Play` instruction
- ``d``: :class:`~qiskit.pulse.instructions.Delay` instruction
- ``f``: :class:`~qiskit.pulse.instructions.SetFrequency` instruction
- ``g``: :class:`~qiskit.pulse.instructions.ShiftFrequency` instruction
- ``q``: :class:`~qiskit.pulse.instructions.SetPhase` instruction
- ``r``: :class:`~qiskit.pulse.instructions.ShiftPhase` instruction
- ``b``: :class:`~qiskit.pulse.instructions.RelativeBarrier` instruction
- ``t``: :class:`~qiskit.pulse.instructions.TimeBlockade` instruction
- ``y``: :class:`~qiskit.pulse.instructions.Reference` instruction (new in version 0.7)

.. _qpy_schedule_operands:

SCHEDULE_BLOCK_OPERANDS
~~~~~~~~~~~~~~~~~~~~~~~

The operands of these instances can be serialized through the standard QPY value serialization
mechanism, however there are special object types that only appear in the schedule operands.
Since the operands are serialized as :ref:`qpy_sequence`, each element must be packed with the
INSTRUCTION_PARAM pack struct, where each payload starts with a header block consists of
the char ``type`` and uint64_t ``size``.
Special objects start with the following type key:

- ``c``: :class:`~qiskit.pulse.channels.Channel`
- ``w``: :class:`~qiskit.pulse.library.Waveform`
- ``s``: :class:`~qiskit.pulse.library.SymbolicPulse`
- ``o``: string (operand string, new in version 0.7)

.. _qpy_schedule_channel:

CHANNEL
~~~~~~~

Channel block starts with channel subtype ``char`` that maps an object data to
:class:`~qiskit.pulse.channels.Channel` subclass. Mapping is defined as follows:

- ``d``: :class:`~qiskit.pulse.channels.DriveChannel`
- ``c``: :class:`~qiskit.pulse.channels.ControlChannel`
- ``m``: :class:`~qiskit.pulse.channels.MeasureChannel`
- ``a``: :class:`~qiskit.pulse.channels.AcquireChannel`
- ``e``: :class:`~qiskit.pulse.channels.MemorySlot`
- ``r``: :class:`~qiskit.pulse.channels.RegisterSlot`

The key is immediately followed by the channel index serialized as the INSTRUCTION_PARAM.

.. _qpy_schedule_waveform:

Waveform
~~~~~~~~

Waveform block starts with WAVEFORM header:

.. code-block:: c

    struct {
        double epsilon;
        uint32_t data_size;
        _bool amp_limited;
    }

which is followed by ``data_size`` bytes of complex ``ndarray`` binary generated by numpy.save_.
This represents the complex IQ data points played on a quantum device.
:attr:`~qiskit.pulse.library.Waveform.name` is saved after the samples in the
INSTRUCTION_PARAM pack struct, which can be string or ``None``.

.. _numpy.save: https://numpy.org/doc/stable/reference/generated/numpy.save.html

.. _qpy_schedule_symbolic_pulse:

SymbolicPulse
~~~~~~~~~~~~~

SymbolicPulse block starts with SYMBOLIC_PULSE header:

.. code-block:: c

    struct {
        uint16_t type_size;
        uint16_t envelope_size;
        uint16_t constraints_size;
        uint16_t valid_amp_conditions_size;
        _bool amp_limited;
    }

which is followed by ``type_size`` utf8 bytes of :attr:`.SymbolicPulse.pulse_type` string
that represents a class of waveform, such as "Gaussian" or "GaussianSquare".
Then, ``envelope_size``, ``constraints_size``, ``valid_amp_conditions_size`` utf8 bytes of
serialized symbolic expressions are generated for :attr:`.SymbolicPulse.envelope`,
:attr:`.SymbolicPulse.constraints`, and :attr:`.SymbolicPulse.valid_amp_conditions`, respectively.
Since string representation of these expressions are usually lengthy,
the expression binary is generated by the python zlib_ module with data compression.

To uniquely specify a pulse instance, we also need to store the associated parameters,
which consist of ``duration`` and the rest of parameters as a dictionary.
Dictionary parameters are first dumped in the :ref:`qpy_mapping` form, and then ``duration``
is dumped with the INSTRUCTION_PARAM pack struct.
Lastly, :attr:`~qiskit.pulse.library.SymbolicPulse.name` is saved also with the
INSTRUCTION_PARAM pack struct, which can be string or ``None``.

.. _zlib: https://docs.python.org/3/library/zlib.html

.. _qpy_mapping:

MAPPING
~~~~~~~

The MAPPING is a representation for arbitrary mapping object. This is a fixed length
:ref:`qpy_sequence` of key-value pair represented by the MAP_ITEM payload.

A MAP_ITEM starts with a header defined as:

.. code-block:: c

    struct {
        uint16_t key_size;
        char type;
        uint16_t size;
    }

which is immediately followed by the ``key_size`` utf8 bytes representing
the dictionary key in string and ``size`` utf8 bytes of arbitrary object data of
QPY serializable ``type``.

.. _qpy_circuit_calibrations:

CIRCUIT_CALIBRATIONS
~~~~~~~~~~~~~~~~~~~~

The CIRCUIT_CALIBRATIONS block is a dictionary to define pulse calibrations of the custom
instruction set. This block starts with the following CALIBRATION header:

.. code-block:: c

    struct {
        uint16_t num_cals;
    }

which is followed by the ``num_cals`` length of calibration entries, each starts with
the CALIBRATION_DEF header:

.. code-block:: c

    struct {
        uint16_t name_size;
        uint16_t num_qubits;
        uint16_t num_params;
        char type;
    }

The calibration definition header is then followed by ``name_size`` utf8 bytes of
the gate name, ``num_qubits`` length of integers representing a sequence of qubits,
and ``num_params`` length of INSTRUCTION_PARAM payload for parameters
associated to the custom instruction.
The ``type`` indicates the class of pulse program which is either, in principle,
``ScheduleBlock`` or :class:`~.Schedule`. As of QPY Version 5,
only ``ScheduleBlock`` payload is supported.
Finally, :ref:`qpy_schedule_block` payload is packed for each CALIBRATION_DEF entry.

.. _qpy_instruction_v5:

INSTRUCTION
~~~~~~~~~~~

The INSTRUCTION block was modified to add two new fields ``num_ctrl_qubits`` and ``ctrl_state``
which are used to model the :attr:`.ControlledGate.num_ctrl_qubits` and
:attr:`.ControlledGate.ctrl_state` attributes. The new payload packed struct
format is:

.. code-block:: c

    struct {
        uint16_t name_size;
        uint16_t label_size;
        uint16_t num_parameters;
        uint32_t num_qargs;
        uint32_t num_cargs;
        _Bool has_conditional;
        uint16_t conditional_reg_name_size;
        int64_t conditional_value;
        uint32_t num_ctrl_qubits;
        uint32_t ctrl_state;
    }

The rest of the instruction payload is the same. You can refer to
:ref:`qpy_instructions` for the details of the full payload.

CUSTOM_INSTRUCTION
~~~~~~~~~~~~~~~~~~

The CUSTOM_INSTRUCTION block in QPY version 5 adds a new field
``base_gate_size`` which is used to define the size of the
:class:`qiskit.circuit.Instruction` object stored in the
:attr:`.ControlledGate.base_gate` attribute for a custom
:class:`~.ControlledGate` object. With this change the CUSTOM_INSTRUCTION
metadata block becomes:

.. code-block:: c

    struct {
        uint16_t name_size;
        char type;
        uint32_t num_qubits;
        uint32_t num_clbits;
        _Bool custom_definition;
        uint64_t size;
        uint32_t num_ctrl_qubits;
        uint32_t ctrl_state;
        uint64_t base_gate_size
    }

Immediately following the CUSTOM_INSTRUCTION struct is the utf8 encoded name
of size ``name_size``.

If ``custom_definition`` is ``True`` that means that the immediately following
``size`` bytes contains a QPY circuit data which can be used for the custom
definition of that gate. If ``custom_definition`` is ``False`` then the
instruction can be considered opaque (ie no definition). The ``type`` field
determines what type of object will get created with the custom definition.
If it's ``'g'`` it will be a :class:`~qiskit.circuit.Gate` object, ``'i'``
it will be a :class:`~qiskit.circuit.Instruction` object.

Following this the next ``base_gate_size`` bytes contain the ``INSTRUCTION``
payload for the :attr:`.ControlledGate.base_gate`.

Additionally an addition value for ``type`` is added ``'c'`` which is used to
indicate the custom instruction is a custom :class:`~.ControlledGate`.

.. _qpy_version_4:

Version 4
---------

Version 4 is identical to :ref:`qpy_version_3` except that it adds 2 new type strings
to the INSTRUCTION_PARAM struct, ``z`` to represent ``None`` (which is encoded as
no data), ``q`` to represent a :class:`.QuantumCircuit` (which is encoded as
a QPY circuit), ``r`` to represent a ``range`` of integers (which is encoded as
a :ref:`qpy_range_pack`), and ``t`` to represent a ``sequence`` (which is encoded as
defined by :ref:`qpy_sequence`). Additionally, version 4 changes the type of register
index mapping array from ``uint32_t`` to ``int64_t``. If the values of any of the
array elements are negative they represent a register bit that is not present in the
circuit.

The :ref:`qpy_registers` header format has also been updated to

.. code-block:: c

    struct {
        char type;
        _Bool standalone;
        uint32_t size;
        uint16_t name_size;
        _bool in_circuit;
    }

which just adds the ``in_circuit`` field which represents whether the register is
part of the circuit or not.

.. _qpy_range_pack:

RANGE
~~~~~

A RANGE is a representation of a ``range`` object. It is defined as:

.. code-block:: c

    struct {
        int64_t start;
        int64_t stop;
        int64_t step;
    }

.. _qpy_sequence:

SEQUENCE
~~~~~~~~

A SEQUENCE is a representation of an arbitrary sequence object. As sequence are just fixed length
containers of arbitrary python objects their QPY can't fully represent any sequence,
but as long as the contents in a sequence are other QPY serializable types for
the INSTRUCTION_PARAM payload the ``sequence`` object can be serialized.

A sequence instruction parameter starts with a header defined as:

.. code-block:: c

    struct {
        uint64_t size;
    }

followed by ``size`` elements that are INSTRUCTION_PARAM payloads, where each of
these define an element in the sequence. The sequence object will be typecasted
into proper type, e.g. ``tuple``, afterwards.

.. _qpy_version_3:

Version 3
---------

Version 3 of the QPY format is identical to :ref:`qpy_version_2` except that it defines
a struct format to represent a :class:`~qiskit.circuit.library.PauliEvolutionGate`
natively in QPY. To accomplish this the :ref:`qpy_custom_definition` struct now supports
a new type value ``'p'`` to represent a :class:`~qiskit.circuit.library.PauliEvolutionGate`.
Enties in the custom instructions tables have unique name generated that start with the
string ``"###PauliEvolutionGate_"`` followed by a uuid string. This gate name is reservered
in QPY and if you have a custom :class:`~qiskit.circuit.Instruction` object with a definition
set and that name prefix it will error. If it's of type ``'p'`` the data payload is defined
as follows:

.. _pauli_evo_qpy:

PAULI_EVOLUTION
~~~~~~~~~~~~~~~

This represents the high level :class:`~qiskit.circuit.library.PauliEvolutionGate`

.. code-block:: c

    struct {
        uint64_t operator_count;
        _Bool standalone_op;
        char time_type;
        uint64_t time_size;
        uint64_t synthesis_size;
    }

This is immediately followed by ``operator_count`` elements defined by the :ref:`qpy_pauli_sum_op`
payload.  Following that we have ``time_size`` bytes representing the ``time`` attribute. If
``standalone_op`` is ``True`` then there must only be a single operator. The
encoding of these bytes is determined by the value of ``time_type``. Possible values of
``time_type`` are ``'f'``, ``'p'``, and ``'e'``. If ``time_type`` is ``'f'`` it's a double,
``'p'`` defines a :class:`~qiskit.circuit.Parameter` object  which is represented by a
:ref:`qpy_param_struct`, ``e`` defines a :class:`~qiskit.circuit.ParameterExpression` object
(that's not a :class:`~qiskit.circuit.Parameter`) which is represented by a :ref:`qpy_param_expr`.
Following that is ``synthesis_size`` bytes which is a utf8 encoded json payload representing
the :class:`.EvolutionSynthesis` class used by the gate.

.. _qpy_pauli_sum_op:

SPARSE_PAULI_OP_LIST_ELEM
~~~~~~~~~~~~~~~~~~~~~~~~~

This represents an instance of :class:`.SparsePauliOp`.


.. code-block:: c

    struct {
        uint32_t pauli_op_size;
    }

which is immediately followed by ``pauli_op_size`` bytes which are .npy format [#f2]_
data which represents the :class:`~qiskit.quantum_info.SparsePauliOp`.

Version 3 of the QPY format also defines a struct format to represent a
:class:`~qiskit.circuit.ParameterVectorElement` as a distinct subclass from
a :class:`~qiskit.circuit.Parameter`. This adds a new parameter type char ``'v'``
to represent a :class:`~qiskit.circuit.ParameterVectorElement` which is now
supported as a type string value for an INSTRUCTION_PARAM. The payload for these
parameters are defined below as :ref:`qpy_param_vector`.

.. _qpy_param_vector:

PARAMETER_VECTOR_ELEMENT
~~~~~~~~~~~~~~~~~~~~~~~~

A PARAMETER_VECTOR_ELEMENT represents a :class:`~qiskit.circuit.ParameterVectorElement`
object the data for a INSTRUCTION_PARAM. The contents of the PARAMETER_VECTOR_ELEMENT are
defined as:

.. code-block:: c

    struct {
        uint16_t vector_name_size;
        uint64_t vector_size;
        char uuid[16];
        uint64_t index;
    }

which is immediately followed by ``vector_name_size`` utf8 bytes representing
the parameter's vector name.

.. _qpy_param_expr_v3:


PARAMETER_EXPR
~~~~~~~~~~~~~~

Additionally, since QPY format version v3 distinguishes between a
:class:`~qiskit.circuit.Parameter` and :class:`~qiskit.circuit.ParameterVectorElement`
the payload for a :class:`~qiskit.circuit.ParameterExpression` needs to be updated
to distinguish between the types. The following is the modified payload format
which is mostly identical to the format in Version 1 and :ref:`qpy_version_2` but just
modifies the ``map_elements`` struct to include a symbol type field.

A PARAMETER_EXPR represents a :class:`~qiskit.circuit.ParameterExpression`
object that the data for an INSTRUCTION_PARAM. The contents of a PARAMETER_EXPR
are defined as:

.. code-block:: c

    struct {
        uint64_t map_elements;
        uint64_t expr_size;
    }

Immediately following the header is ``expr_size`` bytes of utf8 data containing
the expression string, which is the sympy srepr of the expression for the
parameter expression. Following that is a symbol map which contains
``map_elements`` elements with the format

.. code-block:: c

    struct {
        char symbol_type;
        char type;
        uint64_t size;
    }

The ``symbol_type`` key determines the payload type of the symbol representation
for the element. If it's ``p`` it represents a :class:`~qiskit.circuit.Parameter`
and if it's ``v`` it represents a :class:`~qiskit.circuit.ParameterVectorElement`.
The map element struct is immediately followed by the symbol map key payload, if
``symbol_type`` is ``p`` then it is followed immediately by a :ref:`qpy_param_struct`
object (both the struct and utf8 name bytes) and if ``symbol_type`` is ``v``
then the struct is imediately followed by :ref:`qpy_param_vector` (both the struct
and utf8 name bytes). That is followed by ``size`` bytes for the
data of the symbol. The data format is dependent on the value of ``type``. If
``type`` is ``p`` then it represents a :class:`~qiskit.circuit.Parameter` and
size will be 0, the value will just be the same as the key. Similarly if the
``type`` is ``v`` then it represents a :class:`~qiskit.circuit.ParameterVectorElement`
and size will be 0 as the value will just be the same as the key. If
``type`` is ``f`` then it represents a double precision float. If ``type`` is
``c`` it represents a double precision complex, which is represented by the
:ref:`qpy_complex`. Finally, if type is ``i`` it represents an integer which is an
``int64_t``.

.. _qpy_version_2:

Version 2
---------

Version 2 of the QPY format is identical to version 1 except for the HEADER
section is slightly different. You can refer to the :ref:`qpy_version_1` section
for the details on the rest of the payload format.

HEADER
~~~~~~

The contents of HEADER are defined as a C struct are:

.. code-block:: c

    struct {
        uint16_t name_size;
        char global_phase_type;
        uint16_t global_phase_size;
        uint32_t num_qubits;
        uint32_t num_clbits;
        uint64_t metadata_size;
        uint32_t num_registers;
        uint64_t num_instructions;
    }

This is immediately followed by ``name_size`` bytes of utf8 data for the name
of the circuit. Following this is immediately ``global_phase_size`` bytes
representing the global phase. The content of that data is dictated by the
value of ``global_phase_type``. If it's ``'f'`` the data is a float and is the
size of a ``double``. If it's ``'p'`` defines a :class:`~qiskit.circuit.Parameter`
object  which is represented by a PARAM struct (see below), ``e`` defines a
:class:`~qiskit.circuit.ParameterExpression` object (that's not a
:class:`~qiskit.circuit.Parameter`) which is represented by a PARAM_EXPR struct
(see below).

.. _qpy_version_1:

Version 1
---------

HEADER
~~~~~~

The contents of HEADER as defined as a C struct are:

.. code-block:: c

    struct {
        uint16_t name_size;
        double global_phase;
        uint32_t num_qubits;
        uint32_t num_clbits;
        uint64_t metadata_size;
        uint32_t num_registers;
        uint64_t num_instructions;
    }

This is immediately followed by ``name_size`` bytes of utf8 data for the name
of the circuit.

METADATA
~~~~~~~~

The METADATA field is a UTF8 encoded JSON string. After reading the HEADER
(which is a fixed size at the start of the QPY file) and the ``name`` string
you then read the ``metadata_size`` number of bytes and parse the JSON to get
the metadata for the circuit.

.. _qpy_registers:

REGISTERS
~~~~~~~~~

The contents of REGISTERS is a number of REGISTER object. If num_registers is
> 0 then after reading METADATA you read that number of REGISTER structs defined
as:

.. code-block:: c

    struct {
        char type;
        _Bool standalone;
        uint32_t size;
        uint16_t name_size;
    }

``type`` can be ``'q'`` or ``'c'``.

Immediately following the REGISTER struct is the utf8 encoded register name of
size ``name_size``. After the ``name`` utf8 bytes there is then an array of
int64_t values of size ``size`` that contains a map of the register's index to
the circuit's qubit index. For example, array element 0's value is the index
of the ``register[0]``'s position in the containing circuit's qubits list.

.. note::

    Prior to QPY :ref:`qpy_version_4` the type of array elements was uint32_t. This was changed
    to enable negative values which represent bits in the array not present in the
    circuit


The standalone boolean determines whether the register is constructed as a
standalone register that was added to the circuit or was created from existing
bits. A register is considered standalone if it has bits constructed solely
as part of it, for example::

    qr = QuantumRegister(2)
    qc = QuantumCircuit(qr)

the register ``qr`` would be a standalone register. While something like::

    bits = [Qubit(), Qubit()]
    qr2 = QuantumRegister(bits=bits)
    qc = QuantumCircuit(qr2)

``qr2`` would have ``standalone`` set to ``False``.


.. _qpy_custom_definition:

CUSTOM_DEFINITIONS
~~~~~~~~~~~~~~~~~~

This section specifies custom definitions for any of the instructions in the circuit.

CUSTOM_DEFINITION_HEADER contents are defined as:

.. code-block:: c

    struct {
        uint64_t size;
    }

If size is greater than 0 that means the circuit contains custom instruction(s).
Each custom instruction is defined with a CUSTOM_INSTRUCTION block defined as:

.. code-block:: c

    struct {
        uint16_t name_size;
        char type;
        uint32_t num_qubits;
        uint32_t num_clbits;
        _Bool custom_definition;
        uint64_t size;
    }

Immediately following the CUSTOM_INSTRUCTION struct is the utf8 encoded name
of size ``name_size``.

If ``custom_definition`` is ``True`` that means that the immediately following
``size`` bytes contains a QPY circuit data which can be used for the custom
definition of that gate. If ``custom_definition`` is ``False`` then the
instruction can be considered opaque (ie no definition). The ``type`` field
determines what type of object will get created with the custom definition.
If it's ``'g'`` it will be a :class:`~qiskit.circuit.Gate` object, ``'i'``
it will be a :class:`~qiskit.circuit.Instruction` object.

.. _qpy_instructions:

INSTRUCTIONS
~~~~~~~~~~~~

The contents of INSTRUCTIONS is a list of INSTRUCTION metadata objects

.. code-block:: c

    struct {
        uint16_t name_size;
        uint16_t label_size;
        uint16_t num_parameters;
        uint32_t num_qargs;
        uint32_t num_cargs;
        _Bool has_conditional;
        uint16_t conditional_reg_name_size;
        int64_t conditional_value;
    }

This metadata object is immediately followed by ``name_size`` bytes of utf8 bytes
for the ``name``. ``name`` here is the Qiskit class name for the Instruction
class if it's defined in Qiskit. Otherwise it falls back to the custom
instruction name. Following the ``name`` bytes there are ``label_size`` bytes of
utf8 data for the label if one was set on the instruction. Following the label
bytes if ``has_conditional`` is ``True`` then there are
``conditional_reg_name_size`` bytes of utf8 data for the name of the conditional
register name. In case of single classical bit conditions the register name
utf8 data will be prefixed with a null character "\\x00" and then a utf8 string
integer representing the classical bit index in the circuit that the condition
is on.

This is immediately followed by the INSTRUCTION_ARG structs for the list of
arguments of that instruction. These are in the order of all quantum arguments
(there are num_qargs of these) followed by all classical arguments (num_cargs
of these).

The contents of each INSTRUCTION_ARG is:

.. code-block:: c

    struct {
        char type;
        uint32_t index;
    }

``type`` can be ``'q'`` or ``'c'``.

After all arguments for an instruction the parameters are specified with
``num_parameters`` INSTRUCTION_PARAM structs.

The contents of each INSTRUCTION_PARAM is:

.. code-block:: c

    struct {
        char type;
        uint64_t size;
    }

After each INSTRUCTION_PARAM the next ``size`` bytes are the parameter's data.
The ``type`` field can be ``'i'``, ``'f'``, ``'p'``, ``'e'``, ``'s'``, ``'c'``
or ``'n'`` which dictate the format. For ``'i'`` it's an integer, ``'f'`` it's
a double, ``'s'`` if it's a string (encoded as utf8), ``'c'`` is a complex and
the data is represented by the struct format in the :ref:`qpy_param_expr` section.
``'p'`` defines a :class:`~qiskit.circuit.Parameter` object  which is
represented by a :ref:`qpy_param_struct` struct, ``e`` defines a
:class:`~qiskit.circuit.ParameterExpression` object (that's not a
:class:`~qiskit.circuit.Parameter`) which is represented by a :ref:`qpy_param_expr`
struct (on QPY format :ref:`qpy_version_3` the format is tweak slightly see:
:ref:`qpy_param_expr_v3`), ``'n'`` represents an object from numpy (either an
``ndarray`` or a numpy type) which means the data is .npy format [#f2]_ data,
and in QPY :ref:`qpy_version_3` ``'v'`` represents a
:class:`~qiskit.circuit.ParameterVectorElement` which is represented by a
:ref:`qpy_param_vector` struct.

.. _qpy_param_struct:

PARAMETER
~~~~~~~~~

A PARAMETER represents a :class:`~qiskit.circuit.Parameter` object the data for
a INSTRUCTION_PARAM. The contents of the PARAMETER are defined as:

.. code-block:: c

    struct {
        uint16_t name_size;
        char uuid[16];
    }

which is immediately followed by ``name_size`` utf8 bytes representing the
parameter name.

.. _qpy_param_expr:

PARAMETER_EXPR
~~~~~~~~~~~~~~

A PARAMETER_EXPR represents a :class:`~qiskit.circuit.ParameterExpression`
object that the data for an INSTRUCTION_PARAM. The contents of a PARAMETER_EXPR
are defined as:

The PARAMETER_EXPR data starts with a header:

.. code-block:: c

    struct {
        uint64_t map_elements;
        uint64_t expr_size;
    }

Immediately following the header is ``expr_size`` bytes of utf8 data containing
the expression string, which is the sympy srepr of the expression for the
parameter expression. Follwing that is a symbol map which contains
``map_elements`` elements with the format

.. code-block:: c

    struct {
        char type;
        uint64_t size;
    }

Which is followed immediately by ``PARAMETER`` object (both the struct and utf8
name bytes) for the symbol map key. That is followed by ``size`` bytes for the
data of the symbol. The data format is dependent on the value of ``type``. If
``type`` is ``p`` then it represents a :class:`~qiskit.circuit.Parameter` and
size will be 0, the value will just be the same as the key. If
``type`` is ``f`` then it represents a double precision float. If ``type`` is
``c`` it represents a double precision complex, which is represented by :ref:`qpy_complex`.
Finally, if type is ``i`` it represents an integer which is an ``int64_t``.

.. _qpy_complex:

COMPLEX
~~~~~~~

When representing a double precision complex value in QPY the following
struct is used:


.. code-block:: c

    struct {
        double real;
        double imag;
    }

this matches the internal C representation of Python's complex type. [#f3]_

References
==========

.. [#f1] https://tools.ietf.org/html/rfc1700
.. [#f2] https://numpy.org/doc/stable/reference/generated/numpy.lib.format.html
.. [#f3] https://docs.python.org/3/c-api/complex.html#c.Py_complex
"""

from .exceptions import QpyError, UnsupportedFeatureForVersion, QPYLoadingDeprecatedFeatureWarning
from .interface import dump, load

# For backward compatibility. Provide, Runtime, Experiment call these private functions.
from .binary_io import (
    _write_instruction,
    _read_instruction,
    _write_parameter_expression,
    _read_parameter_expression,
    _read_parameter_expression_v3,
)
from .common import QPY_VERSION, QPY_COMPATIBILITY_VERSION



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""The Boolean logic circuit library."""

from .quantum_and import AND, AndGate
from .quantum_or import OR, OrGate
from .quantum_xor import XOR, BitwiseXorGate, random_bitwise_xor
from .inner_product import InnerProduct, InnerProductGate



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

""" Test of GenericBackendV2 backend"""

import math
import operator
import unittest
from ddt import ddt, data

from qiskit import ClassicalRegister, QuantumCircuit, QuantumRegister, transpile
from qiskit.providers.fake_provider import GenericBackendV2
from qiskit.quantum_info import Operator
from qiskit.transpiler import CouplingMap
from qiskit.utils import optionals
from qiskit.exceptions import QiskitError
from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
from test import QiskitTestCase, combine  # pylint: disable=wrong-import-order


BACKENDS = []
for n in [5, 7, 16, 20, 27, 65, 127]:
    cmap = CouplingMap.from_ring(n)
    BACKENDS.append(GenericBackendV2(num_qubits=n, coupling_map=cmap, seed=42))


@ddt
class TestGenericBackendV2(QiskitTestCase):
    """Test class for GenericBackendV2 backend"""

    def setUp(self):
        super().setUp()
        self.cmap = CouplingMap(
            [(0, 2), (0, 1), (1, 3), (2, 4), (2, 3), (3, 5), (4, 6), (4, 5), (5, 7), (6, 7)]
        )

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.circuit = QuantumCircuit(2)
        cls.circuit.h(0)
        cls.circuit.h(1)
        cls.circuit.h(0)
        cls.circuit.h(1)
        cls.circuit.x(0)
        cls.circuit.x(1)
        cls.circuit.measure_all()

    def test_supported_basis_gates(self):
        """Test that target raises error if basis_gate not in ``supported_names``."""
        with self.assertRaises(QiskitError):
            GenericBackendV2(num_qubits=8, basis_gates=["cx", "id", "rz", "sx", "zz"], seed=42)

    def test_cx_1Q(self):
        """Test failing with a backend with single qubit but with a two-qubit basis gate"""
        with self.assertRaises(QiskitError):
            GenericBackendV2(num_qubits=1, basis_gates=["cx", "id"], seed=42)

    def test_ccx_2Q(self):
        """Test failing with a backend with two qubits but with a three-qubit basis gate"""
        with self.assertRaises(QiskitError):
            GenericBackendV2(num_qubits=2, basis_gates=["ccx", "id"], seed=42)

    def test_no_noise(self):
        """Test no noise info when parameter is false"""
        backend = GenericBackendV2(
            num_qubits=5, coupling_map=CouplingMap.from_line(5), noise_info=False, seed=42
        )
        qc = QuantumCircuit(5)
        qc.h(0)
        qc.cx(0, 1)
        qc.cx(0, 2)
        qc.cx(1, 4)
        qc.cx(3, 0)
        qc.cx(2, 4)
        qc_res = generate_preset_pass_manager(optimization_level=2, backend=backend).run(qc)
        self.assertTrue(Operator.from_circuit(qc_res).equiv(qc))
        self.assertEqual(backend.target.qubit_properties, None)

    def test_no_noise_fully_connected(self):
        """Test no noise info when parameter is false"""
        backend = GenericBackendV2(num_qubits=5, noise_info=False, seed=42)
        qc = QuantumCircuit(5)
        qc.h(0)
        qc.cx(0, 1)
        qc.cx(0, 2)
        qc.cx(1, 4)
        qc.cx(3, 0)
        qc.cx(2, 4)
        qc_res = generate_preset_pass_manager(optimization_level=2, backend=backend).run(qc)
        self.assertTrue(Operator.from_circuit(qc_res).equiv(qc))
        self.assertEqual(backend.target.qubit_properties, None)

    def test_no_info(self):
        """Test no noise info when parameter is false"""
        backend = GenericBackendV2(
            num_qubits=5,
            coupling_map=CouplingMap.from_line(5),
            noise_info=False,
            seed=42,
        )
        qc = QuantumCircuit(5)
        qc.h(0)
        qc.cx(0, 1)
        qc.cx(0, 2)
        qc.cx(1, 4)
        qc.cx(3, 0)
        qc.cx(2, 4)
        qc_res = generate_preset_pass_manager(optimization_level=2, backend=backend).run(qc)
        self.assertTrue(Operator.from_circuit(qc_res).equiv(qc))
        self.assertEqual(backend.target.qubit_properties, None)

    def test_operation_names(self):
        """Test that target basis gates include "delay", "measure" and "reset" even
        if not provided by user."""
        target = GenericBackendV2(num_qubits=8, seed=42)
        op_names = list(target.operation_names)
        op_names.sort()
        self.assertEqual(op_names, ["cx", "delay", "id", "measure", "reset", "rz", "sx", "x"])

        target = GenericBackendV2(num_qubits=8, basis_gates=["ecr", "id", "rz", "sx", "x"], seed=42)
        op_names = list(target.operation_names)
        op_names.sort()
        self.assertEqual(op_names, ["delay", "ecr", "id", "measure", "reset", "rz", "sx", "x"])

    def test_incompatible_coupling_map(self):
        """Test that the size of the coupling map must match num_qubits."""
        with self.assertRaises(QiskitError):
            GenericBackendV2(num_qubits=5, coupling_map=self.cmap, seed=42)

    def test_control_flow_operation_names(self):
        """Test that control flow instructions are added to the target if control_flow is True."""
        target = GenericBackendV2(
            num_qubits=8,
            basis_gates=["ecr", "id", "rz", "sx", "x"],
            coupling_map=self.cmap,
            control_flow=True,
            seed=42,
        ).target
        op_names = list(target.operation_names)
        op_names.sort()
        reference = [
            "break",
            "continue",
            "delay",
            "ecr",
            "for_loop",
            "id",
            "if_else",
            "measure",
            "reset",
            "rz",
            "switch_case",
            "sx",
            "while_loop",
            "x",
        ]
        self.assertEqual(op_names, reference)

    def test_default_coupling_map(self):
        """Test that fully-connected coupling map is generated correctly."""

        # fmt: off
        reference_cmap = [(0, 1), (1, 0), (0, 2), (2, 0), (0, 3), (3, 0), (0, 4), (4, 0), (1, 2), (2, 1),
                          (1, 3), (3, 1), (1, 4), (4, 1), (2, 3), (3, 2), (2, 4), (4, 2), (3, 4), (4, 3)]
        # fmt: on
        self.assertEqual(
            list(GenericBackendV2(num_qubits=5, seed=42).coupling_map.get_edges()),
            reference_cmap,
        )

    def test_run(self):
        """Test run method, confirm correct noisy simulation if Aer is installed."""

        qr = QuantumRegister(5)
        cr = ClassicalRegister(5)
        qc = QuantumCircuit(qr, cr)
        qc.h(qr[0])
        for k in range(1, 4):
            qc.cx(qr[0], qr[k])
        qc.measure(qr, cr)

        backend = GenericBackendV2(num_qubits=5, basis_gates=["cx", "id", "rz", "sx", "x"], seed=42)
        tqc = transpile(qc, backend=backend, optimization_level=3, seed_transpiler=42)
        result = backend.run(tqc, seed_simulator=42, shots=1000).result()
        counts = result.get_counts()

        self.assertTrue(math.isclose(counts["00000"], 500, rel_tol=0.1))
        self.assertTrue(math.isclose(counts["01111"], 500, rel_tol=0.1))

    def test_duration_defaults(self):
        """Test that the basis gates are assigned duration defaults within expected ranges."""

        basis_gates = ["cx", "id", "rz", "sx", "x", "sdg", "rxx"]
        expected_durations = {
            "cx": (7.992e-08, 8.99988e-07),
            "id": (2.997e-08, 5.994e-08),
            "rz": (0.0, 0.0),
            "sx": (2.997e-08, 5.994e-08),
            "x": (2.997e-08, 5.994e-08),
            "measure": (6.99966e-07, 1.500054e-06),
            "sdg": (2.997e-08, 5.994e-08),
            "rxx": (7.992e-08, 8.99988e-07),
        }
        for _ in range(20):
            target = GenericBackendV2(num_qubits=2, basis_gates=basis_gates, seed=42).target
            for inst in target:
                for qargs in target.qargs_for_operation_name(inst):
                    duration = target[inst][qargs].duration
                    if inst not in ["delay", "reset"]:
                        self.assertGreaterEqual(duration, expected_durations[inst][0])
                        self.assertLessEqual(duration, expected_durations[inst][1])

    def test_custom_dt(self):
        """Test that the custom dt is respected."""

        ref_backend = GenericBackendV2(num_qubits=2, basis_gates=["cx", "id"], seed=42)
        double_dt_backend = GenericBackendV2(
            num_qubits=2, basis_gates=["cx", "id"], dt=ref_backend.dt * 2, seed=42
        )
        self.assertEqual(ref_backend.dt * 2, double_dt_backend.dt)

    @combine(
        backend=BACKENDS,
        optimization_level=[0, 1, 2, 3],
    )
    def test_circuit_on_generic_backend_v2(self, backend, optimization_level):
        """Test run method."""

        if not optionals.HAS_AER and backend.num_qubits > 20:
            self.skipTest(f"Unable to run generic backend {backend.name} without qiskit-aer")
        job = backend.run(
            transpile(
                self.circuit, backend, seed_transpiler=42, optimization_level=optimization_level
            ),
            seed_simulator=42,
        )
        result = job.result()
        counts = result.get_counts()
        max_count = max(counts.items(), key=operator.itemgetter(1))[0]
        self.assertEqual(max_count, "11")

    @data(*BACKENDS)
    def test_backend_v2_dt(self, backend):
        """Test default dt value is consistent with legacy fake backends."""

        target = backend.target
        if target.dt is not None:
            self.assertLess(target.dt, 1e-6)

    @data(*BACKENDS)
    def test_backend_v2_dtm(self, backend):
        """Test default dtm value is consistent with legacy fake backends."""

        if backend.dtm:
            self.assertLess(backend.dtm, 1e-6)

    @unittest.skipUnless(optionals.HAS_AER, "qiskit-aer is required to run this test")
    def test_generic_noise_model_always_present(self):
        """Test that GenericBackendV2 instances run with noise if Aer installed."""

        backend = GenericBackendV2(num_qubits=5, seed=42)
        backend.set_options(seed_simulator=42)
        qc = QuantumCircuit(1)
        qc.x(0)
        qc.measure_all()

        res = backend.run(qc, shots=1000).result().get_counts()
        # Assert noise was present and result wasn't ideal
        self.assertNotEqual(res, {"1": 1000})



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Unify time unit in circuit for scheduling and following passes."""
from typing import Set

from qiskit.circuit import Delay, Duration
from qiskit.circuit.classical import expr
from qiskit.circuit.duration import duration_in_dt
from qiskit.dagcircuit import DAGCircuit
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.exceptions import TranspilerError
from qiskit.transpiler.instruction_durations import InstructionDurations
from qiskit.transpiler.target import Target
from qiskit.utils import apply_prefix


class TimeUnitConversion(TransformationPass):
    """Choose a time unit to be used in the following time-aware passes,
    and make all circuit time units consistent with that.

    This pass will add a :attr:`.Instruction.duration` metadata to each op whose duration is known
    which will be used by subsequent scheduling passes for scheduling.

    If ``dt`` (in seconds) is known to transpiler, the unit ``'dt'`` is chosen. Otherwise,
    the unit to be selected depends on what units are used in delays and instruction durations:

    * ``'s'``: if they are all in SI units.
    * ``'dt'``: if they are all in the unit ``'dt'``.
    * raise error: if they are a mix of SI units and ``'dt'``.
    """

    def __init__(self, inst_durations: InstructionDurations = None, target: Target = None):
        """TimeUnitAnalysis initializer.

        Args:
            inst_durations (InstructionDurations): A dictionary of durations of instructions.
            target: The :class:`~.Target` representing the target backend, if both
                  ``inst_durations`` and ``target`` are specified then this argument will take
                  precedence and ``inst_durations`` will be ignored.


        """
        super().__init__()
        self.inst_durations = inst_durations or InstructionDurations()
        if target is not None:
            # The priority order for instruction durations is: target > standalone.
            self.inst_durations = target.durations()
        self._durations_provided = inst_durations is not None or target is not None

    def run(self, dag: DAGCircuit):
        """Run the TimeUnitAnalysis pass on `dag`.

        Args:
            dag (DAGCircuit): DAG to be checked.

        Returns:
            DAGCircuit: DAG with consistent timing and op nodes annotated with duration.

        Raises:
            TranspilerError: if the units are not unifiable
        """

        inst_durations = InstructionDurations()
        if self._durations_provided:
            inst_durations.update(self.inst_durations, getattr(self.inst_durations, "dt", None))

        # The float-value converted units for delay expressions, either all in 'dt'
        # or all in seconds.
        expression_durations = {}

        # Choose unit
        has_dt = False
        has_si = False

        # We _always_ need to traverse duration expressions to convert them to
        # a float. But we also use the opportunity to note if they intermix cycles
        # and wall-time, in case we don't have a `dt` to use to unify all instruction
        # durations.
        for node in dag.op_nodes(op=Delay):
            if isinstance(node.op.duration, expr.Expr):
                if any(
                    isinstance(x, expr.Stretch) for x in expr.iter_identifiers(node.op.duration)
                ):
                    # If any of the delays use a stretch expression, we can't run scheduling
                    # passes anyway, so we bail out. In theory, we _could_ still traverse
                    # through the stretch expression and replace any Duration value nodes it may
                    # contain with ones of the same units, but it'd be complex and probably unuseful.
                    self.property_set["time_unit"] = "stretch"
                    return dag

                visitor = _EvalDurationImpl(inst_durations.dt)
                duration = node.op.duration.accept(visitor)
                if visitor.in_cycles():
                    has_dt = True
                    # We need to round in case the expression evaluated to a non-integral 'dt'.
                    duration = duration_in_dt(duration, 1.0)
                else:
                    has_si = True
                if duration < 0:
                    raise TranspilerError(
                        f"Expression '{node.op.duration}' resolves to a negative duration!"
                    )
                expression_durations[node._node_id] = duration
            else:
                if node.op.unit == "dt":
                    has_dt = True
                else:
                    has_si = True
            if inst_durations.dt is None and has_dt and has_si:
                raise TranspilerError(
                    "Fail to unify time units in delays. SI units "
                    "and dt unit must not be mixed when dt is not supplied."
                )

        if inst_durations.dt is None:
            # Check what units are used in other instructions: dt or SI or mixed
            units_other = inst_durations.units_used()
            unified_unit = self._unified(units_other)
            if unified_unit == "SI" and not has_dt:
                time_unit = "s"
            elif unified_unit == "dt" and not has_si:
                time_unit = "dt"
            else:
                raise TranspilerError(
                    "Fail to unify time units. SI units "
                    "and dt unit must not be mixed when dt is not supplied."
                )
        else:
            time_unit = "dt"

        # Make instructions with local durations consistent.
        for node in dag.op_nodes(Delay):
            op = node.op.to_mutable()
            if node._node_id in expression_durations:
                op.duration = expression_durations[node._node_id]
            else:
                op.duration = inst_durations._convert_unit(op.duration, op.unit, time_unit)
            op.unit = time_unit
            dag.substitute_node(node, op)

        self.property_set["time_unit"] = time_unit
        return dag

    @staticmethod
    def _unified(unit_set: Set[str]) -> str:
        if not unit_set:
            return "dt"

        if len(unit_set) == 1 and "dt" in unit_set:
            return "dt"

        all_si = True
        for unit in unit_set:
            if not unit.endswith("s"):
                all_si = False
                break

        if all_si:
            return "SI"

        return "mixed"


class _EvalDurationImpl(expr.ExprVisitor[float]):
    """Evaluates the expression to a single float result.

    If ``dt`` is provided or all durations are already in ``dt``, the result is in ``dt``.
    Otherwise, the result will be in seconds, and all durations MUST be in wall-time (SI).
    """

    __slots__ = ("dt", "has_dt", "has_si")

    def __init__(self, dt=None):
        self.dt = dt
        self.has_dt = False
        self.has_si = False

    def in_cycles(self):
        """Returns ``True`` if units are 'dt' after visit."""
        return self.has_dt or self.dt is not None

    def visit_value(self, node, /) -> float:
        if isinstance(node.value, float):
            return node.value
        if isinstance(node.value, Duration.dt):
            if self.has_si and self.dt is None:
                raise TranspilerError(
                    "Fail to unify time units in delays. SI units "
                    "and dt unit must not be mixed when dt is not supplied."
                )
            self.has_dt = True
            return node.value[0]
        if isinstance(node.value, Duration):
            if self.has_dt and self.dt is None:
                raise TranspilerError(
                    "Fail to unify time units in delays. SI units "
                    "and dt unit must not be mixed when dt is not supplied."
                )
            self.has_si = True
            # Setting 'divisor' to 1 when there's no 'dt' is just to simplify
            # the logic (we don't need to divide).
            divisor = self.dt if self.dt is not None else 1
            if isinstance(node.value, Duration.s):
                return node.value[0] / divisor
            from_unit = node.value.unit()
            return apply_prefix(node.value[0], from_unit) / divisor
        raise TranspilerError(f"invalid duration expression: {node}")

    def visit_binary(self, node, /) -> float:
        left = node.left.accept(self)
        right = node.right.accept(self)
        if node.op == expr.Binary.Op.ADD:
            return left + right
        if node.op == expr.Binary.Op.SUB:
            return left - right
        if node.op == expr.Binary.Op.MUL:
            return left * right
        if node.op == expr.Binary.Op.DIV:
            return left / right
        raise TranspilerError(f"invalid duration expression: {node}")

    def visit_cast(self, node, /) -> float:
        return node.operand.accept(self)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Parser for sympy expressions srepr from ParameterExpression internals."""

import ast

from qiskit.qpy.exceptions import QpyError


ALLOWED_CALLERS = {
    "Abs",
    "Add",
    "Sub",
    "Mul",
    "Div",
    "Pow",
    "Symbol",
    "Integer",
    "Rational",
    "Complex",
    "Float",
    "log",
    "sin",
    "cos",
    "tan",
    "atan",
    "acos",
    "asin",
    "exp",
    "conjugate",
}

UNARY = {
    "sin",
    "cos",
    "tan",
    "atan",
    "acos",
    "asin",
    "conjugate",
    "Symbol",
    "Integer",
    "Complex",
    "Abs",
    "Float",
}


class ParseSympyWalker(ast.NodeVisitor):
    """A custom ast walker that is passed the sympy srepr from QPY < 13 and creates a custom
    expression."""

    def __init__(self):
        self.stack = []

    def visit_UnaryOp(self, node: ast.UnaryOp):  # pylint: disable=invalid-name
        """Visit a python unary op node"""
        self.visit(node.operand)
        arg = self.stack.pop()
        if isinstance(node.op, ast.UAdd):
            self.stack.append(+arg)
        elif isinstance(node.op, ast.USub):
            self.stack.append(-arg)
        elif isinstance(node.op, ast.Not):
            self.stack.append(not arg)
        elif isinstance(node.op, ast.Invert):
            self.stack.append(~arg)
        else:
            raise QpyError(f"Invalid unary op as part of sympy srepr: {node.op}")

    def visit_Constant(self, node: ast.Constant):  # pylint: disable=invalid-name
        """Visit a constant node."""
        self.stack.append(node.value)

    def visit_Call(self, node: ast.Call):  # pylint: disable=invalid-name
        """Visit a call node

        This can only be parameter expression allowed sympy call types.
        """
        import sympy

        if isinstance(node.func, ast.Name):
            name = node.func.id
        else:
            raise QpyError("Unknown node type")

        if name not in ALLOWED_CALLERS:
            raise QpyError(f"{name} is not part of a valid sympy expression srepr")

        args = node.args
        if name in UNARY:
            if len(args) != 1:
                raise QpyError(f"{name} has an invalid number of args in sympy srepr")
            self.visit(args[0])
            obj = getattr(sympy, name)(self.stack.pop())
            self.stack.append(obj)
        else:
            for arg in args:
                self.visit(arg)
            out_args = [self.stack.pop() for _ in range(len(args))]
            out_args.reverse()
            obj = getattr(sympy, name)(*out_args)
            self.stack.append(obj)


def parse_sympy_repr(sympy_repr: str):
    """Parse a given sympy srepr into a symbolic expression object."""
    tree = ast.parse(sympy_repr, mode="eval")
    visitor = ParseSympyWalker()
    visitor.visit(tree)
    return visitor.stack.pop()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Integer comparator based on an exponential number of multi-controlled gates."""

import numpy as np
from qiskit.circuit.quantumcircuit import QuantumCircuit


def synth_integer_comparator_greedy(
    num_state_qubits: int, value: int, geq: bool = True
) -> QuantumCircuit:
    r"""Implement an integer comparison based on value-by-value comparison.

    For ``value`` smaller than ``2 ** (num_state_qubits - 1)`` this circuit implements
    ``value`` multi-controlled gates with control states 0, 1, ..., ``value - 1``, such that
    the target qubit is flipped if the qubit state represents any of the allowed values.
    For ``value`` larger than that, ``geq`` is flipped. This implementation can
    require an exponential number of gates. If auxiliary qubits are available, the implementation
    provided by :func:`.synth_integer_comparator_2s` is more efficient.

    Args:
        num_state_qubits: The number of qubits encoding the value to compare to.
        value: The value to compare to.
        geq: If ``True`` flip the target bit if the qubit state is :math:`\geq` than the value,
            otherwise implement :math:`<`.

    Returns:
        A circuit implementing the integer comparator.

    """
    circuit = QuantumCircuit(num_state_qubits + 1)

    if value <= 0:  # condition always satisfied for non-positive values
        if geq:  # otherwise the condition is never satisfied
            circuit.x(num_state_qubits)

        return circuit

    # make sure to always choose the comparison where we have to place less than
    # (2 ** n)/2 MCX gates
    value = int(np.ceil(value))
    if (value < 2 ** (num_state_qubits - 1) and geq) or (
        value > 2 ** (num_state_qubits - 1) and not geq
    ):
        geq = not geq
        circuit.x(num_state_qubits)

    if geq:
        accepted_values = range(value, 2**num_state_qubits)
    else:
        accepted_values = range(0, value)

    for accepted_value in accepted_values:
        circuit.mcx(list(range(num_state_qubits)), num_state_qubits, ctrl_state=accepted_value)

    return circuit



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Simple box basic block."""

from __future__ import annotations

import typing

from qiskit.circuit.exceptions import CircuitError
from .control_flow import ControlFlowOp

if typing.TYPE_CHECKING:
    from qiskit.circuit import QuantumCircuit


class BoxOp(ControlFlowOp):
    """A scoped "box" of operations on a circuit that are treated atomically in the greater context.

    A "box" is a control-flow construct that is entered unconditionally.  The contents of the box
    behave somewhat as if the start and end of the box were barriers, except it is permissible to
    commute operations "all the way" through the box.  The box is also an explicit scope for the
    purposes of variables, stretches and compiler passes.

    Typically you create this by using the builder-interface form of :meth:`.QuantumCircuit.box`.
    """

    def __init__(
        self,
        body: QuantumCircuit,
        duration: None = None,
        unit: typing.Literal["dt", "s", "ms", "us", "ns", "ps"] = "dt",
        label: str | None = None,
    ):
        """
        Default constructor of :class:`BoxOp`.

        Args:
            body: the circuit to use as the body of the box.  This should explicit close over any
                :class:`.expr.Var` variables that must be incident from the outer circuit.  The
                expected number of qubit and clbits for the resulting instruction are inferred from
                the number in the circuit, even if they are idle.
            duration: an optional duration for the box as a whole.
            unit: the unit of the ``duration``.
            label: an optional string label for the instruction.
        """
        super().__init__("box", body.num_qubits, body.num_clbits, [body], label=label)
        self.duration = duration
        self.unit = unit

    @property
    def params(self):
        return self._params

    @params.setter
    def params(self, parameters):
        # pylint: disable=cyclic-import
        from qiskit.circuit import QuantumCircuit

        (body,) = parameters

        if not isinstance(body, QuantumCircuit):
            raise CircuitError(
                "BoxOp expects a body parameter of type "
                f"QuantumCircuit, but received {type(body)}."
            )

        if body.num_qubits != self.num_qubits or body.num_clbits != self.num_clbits:
            raise CircuitError(
                "Attempted to assign a body parameter with a num_qubits or "
                "num_clbits different than that of the BoxOp. "
                f"BoxOp num_qubits/clbits: {self.num_qubits}/{self.num_clbits} "
                f"Supplied body num_qubits/clbits: {body.num_qubits}/{body.num_clbits}."
            )

        self._params = [body]

    @property
    def body(self):
        """The ``body`` :class:`.QuantumCircuit` of the operation.

        This is the same as object returned as the sole entry in :meth:`params` and :meth:`blocks`.
        """
        # Not settable via this property; the only meaningful way to replace a body is via
        # larger `QuantumCircuit` methods, or using `replace_blocks`.
        return self.params[0]

    @property
    def blocks(self):
        return (self._params[0],)

    def replace_blocks(self, blocks):
        (body,) = blocks
        return BoxOp(body, duration=self.duration, unit=self.unit, label=self.label)

    def __eq__(self, other):
        return (
            isinstance(other, BoxOp)
            and self.duration == other.duration
            and self.unit == other.unit
            and super().__eq__(other)
        )


class BoxContext:
    """Context-manager that powers :meth:`.QuantumCircuit.box`.

    This is not part of the public interface, and should not be instantiated by users.
    """

    __slots__ = ("_circuit", "_duration", "_unit", "_label")

    def __init__(
        self,
        circuit: QuantumCircuit,
        *,
        duration: None = None,
        unit: typing.Literal["dt", "s", "ms", "us", "ns", "ps"] = "dt",
        label: str | None = None,
    ):
        """
        Args:
            circuit: the outermost scope of the circuit under construction.
            duration: the final duration of the box.
            unit: the unit of ``duration``.
            label: an optional label for the box.
        """
        self._circuit = circuit
        self._duration = duration
        self._unit = unit
        self._label = label

    def __enter__(self):
        # For a box to have the semantics of internal qubit alignment with a resolvable duration, we
        # can't allow conditional jumps to exit it.  Technically an unconditional `break` or
        # `continue` could work, but we're not getting into that.
        self._circuit._push_scope(allow_jumps=False)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            # If we're leaving the context manager because an exception was raised, there's nothing
            # to do except restore the circuit state.
            self._circuit._pop_scope()
            return False
        scope = self._circuit._pop_scope()
        # Boxes do not need to pass any further resources in, because there's no jumps out of a
        # `box` permitted.
        body = scope.build(scope.qubits(), scope.clbits())
        self._circuit.append(
            BoxOp(body, duration=self._duration, unit=self._unit, label=self._label),
            body.qubits,
            body.clbits,
        )
        return False



# This code is part of Qiskit.
#
# (C) Copyright IBM 2025.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Circuit synthesizers and related classes for boolean expressions"""

import itertools
from qiskit.circuit import QuantumCircuit
from qiskit.circuit.library import ZGate, XGate


class EsopGenerator:
    """Generates an ESOP (Exlusive-sum-of-products) representation
    for a boolean function given by its truth table"""

    def __init__(self, truth_table):
        self.truth_table = truth_table
        self.cofactor_table = {}
        self.esop = self.generate_esop(assignment=())

    def clause_diff(self, clause1, clause2):
        """The indices of variables where the clauses differ"""
        return [i for i in range(len(clause1)) if clause1[i] != clause2[i]]

    def combine_and_optimize(self, clauses_1, clauses_2):
        """Combining clauses of distance 1 until no more combinations can be performed"""
        something_changed = True
        combined_clauses = clauses_1 + clauses_2
        while something_changed:
            something_changed = False
            for clause_1, clause_2 in itertools.combinations(combined_clauses, 2):
                diff = self.clause_diff(clause_1, clause_2)
                if len(diff) == 1:
                    new_clause = clause_1[: diff[0]] + "-" + clause_1[diff[0] + 1 :]
                    if new_clause not in combined_clauses:
                        combined_clauses.remove(clause_1)
                        combined_clauses.remove(clause_2)
                        combined_clauses.append(new_clause)
                        something_changed = True
                        break
        return combined_clauses

    def generate_esop(self, assignment):
        """Recursively generates an ESOP for a partially determined boolean expression
        'assignment' is a partial assignment to the expression's variables
        """
        # this base case is non-optimized;
        # we may be able to terminate even before having a full assignment
        if len(assignment) == self.truth_table.num_bits:
            if self.truth_table[assignment]:
                return ["-" * self.truth_table.num_bits]  # True; a don't care clause
            else:
                return []  # False
        i = len(assignment)  # current variable
        negative_esop = self.generate_esop(assignment + (False,))
        positive_esop = self.generate_esop(assignment + (True,))
        # Shannon's expansion
        negative_esop_bool_expanded = [
            clause[:i] + "0" + clause[i + 1 :] for clause in negative_esop
        ]
        positive_esop_bool_expanded = [
            clause[:i] + "1" + clause[i + 1 :] for clause in positive_esop
        ]
        self.cofactor_table[assignment] = self.combine_and_optimize(
            negative_esop_bool_expanded, positive_esop_bool_expanded
        )
        return self.cofactor_table[assignment]


def synth_phase_oracle_from_esop(esop, num_qubits):
    """
    Generates a phase oracle for the boolean function f given in ESOP (Exlusive sum of products) form
    esop is of the form ('01-1', '11-0', ...) etc
    where 1 is the variable, 0 is negated variable and - is don't care
    """
    qc = QuantumCircuit(num_qubits)
    clause_data = [
        (zip(*[qubit_data for qubit_data in enumerate(clause) if qubit_data[1] != "-"]))
        for clause in esop
    ]
    for qubit_indices, control_data in clause_data:
        control_state = "".join(control_data)
        if len(control_state) == 1:  # single qubit; either Z or XZX
            if control_state == "0":
                qc.x(qubit_indices[0])
            qc.z(qubit_indices[0])
            if control_state == "0":
                qc.x(qubit_indices[0])
        else:  # use custom controlled-Z gate
            # we use the last qubit as the target, flipping it if the control is 0 for that qubit
            gate = ZGate().control(len(qubit_indices) - 1, ctrl_state=control_state[:-1][::-1])
            if control_state[-1] == "0":
                qc.x(qubit_indices[-1])
            qc.append(gate, qubit_indices)
            if control_state[-1] == "0":
                qc.x(qubit_indices[-1])
    return qc


def synth_bit_oracle_from_esop(esop, num_qubits):
    """
    Generates a bit-flip oracle for the boolean function f given in ESOP (Exlusive sum of products) form
    esop is of the form ('01-1', '11-0', ...) etc
    where 1 is the variable, 0 is negated variable and - is don't care
    """
    output_index = num_qubits - 1
    qc = QuantumCircuit(num_qubits)
    clause_data = [
        (zip(*[qubit_data for qubit_data in enumerate(clause) if qubit_data[1] != "-"]))
        for clause in esop
    ]
    for qubit_indices, control_data in clause_data:
        control_state = "".join(control_data)
        # use custom controlled-X gate
        gate = XGate().control(len(qubit_indices), ctrl_state=control_state[::-1])
        qc.append(gate, qubit_indices + (output_index,))
    return qc
